#!/usr/bin/python3 -B
import argparse
import json
import os
import shutil
import subprocess
import sys
import textwrap
import traceback
from tempfile import TemporaryDirectory
from typing import (
    List,
    Dict,
    Any,
    Tuple,
    Optional,
    NoReturn,
    NamedTuple,
    Literal,
    cast,
)

from debputy import DEBPUTY_ROOT_DIR, DEBPUTY_PLUGIN_ROOT_DIR
from debputy.analysis import REFERENCE_DATA_TABLE
from debputy.analysis.debian_dir import scan_debian_dir
from debputy.build_support import perform_clean, perform_builds
from debputy.commands.debputy_cmd.context import (
    CommandContext,
    add_arg,
    ROOT_COMMAND,
    CommandArg,
)
from debputy.commands.debputy_cmd.output import _stream_to_pager, _output_styling
from debputy.commands.debputy_cmd.plugin_cmds import lookup_pmr_rule
from debputy.dh_migration.migrators import MIGRATORS
from debputy.exceptions import (
    DebputyRuntimeError,
    PluginNotFoundError,
    PluginAPIViolationError,
    PluginInitializationError,
    UnhandledOrUnexpectedErrorFromPluginError,
    SymlinkLoopError,
    DebputyRuntimeErrorWithPreamble,
)
from debputy.highlevel_manifest import HighLevelManifest
from debputy.manifest_parser.parser_doc import render_rule
from debputy.package_build.assemble_deb import (
    assemble_debs,
)
from debputy.plugin.api.spec import (
    INTEGRATION_MODE_DH_DEBPUTY_RRR,
    DebputyIntegrationMode,
    INTEGRATION_MODE_FULL,
)

try:
    from argcomplete import autocomplete
except ImportError:

    def autocomplete(_parser: argparse.ArgumentParser) -> None:
        pass


from debputy.version import __version__
from debputy.filesystem_scan import (
    FSROOverlay,
    FSRootDir,
)
from debputy.plugin.api.impl_types import (
    DebputyPluginMetadata,
)
from debputy.plugin.api.impl import (
    find_json_plugin,
    find_tests_for_plugin,
    find_related_implementation_files_for_plugin,
    parse_json_plugin_desc,
)
from debputy.dh_migration.migration import migrate_from_dh, _check_migration_target
from debputy.dh_migration.models import AcceptableMigrationIssues
from debputy.dh.debhelper_emulation import (
    dhe_pkgdir,
)

from debputy.deb_packaging_support import (
    usr_local_transformation,
    handle_perl_code,
    detect_systemd_user_service_files,
    fixup_debian_changelog_and_news_file,
    install_upstream_changelog,
    relocate_dwarves_into_dbgsym_packages,
    run_package_processors,
    cross_package_control_files,
)
from debputy.util import (
    _error,
    _warn,
    ColorizedArgumentParser,
    setup_logging,
    _info,
    escape_shell,
    program_name,
    integrated_with_debhelper,
    PRINT_BUILD_SYSTEM_COMMAND,
    PRINT_COMMAND,
    manifest_format_doc,
)


class SharedArgument(NamedTuple):
    """
    Information about an argument shared between a parser and its subparsers
    """

    action: argparse.Action
    args: Tuple[Any, ...]
    kwargs: Dict[str, Any]


class Namespace(argparse.Namespace):
    """
    Hacks around a namespace to allow merging of values set multiple times

    Based on: https://www.enricozini.org/blog/2022/python/sharing-argparse-arguments-with-subcommands/
    """

    def __setattr__(self, name: str, value: Any) -> None:
        arg = self._shared_args.get(name)
        if arg is not None:
            action_type = arg.kwargs.get("action")
            if action_type == "store_true":
                # OR values
                old = getattr(self, name, False)
                super().__setattr__(name, old or value)
            elif action_type == "store_false":
                # AND values
                old = getattr(self, name, True)
                super().__setattr__(name, old and value)
            elif action_type == "append":
                old = getattr(self, name, None)
                if old is None:
                    old = []
                    super().__setattr__(name, old)
                if isinstance(value, list):
                    old.extend(value)
                elif value is not None:
                    old.append(value)
            elif action_type == "store":
                old = getattr(self, name, None)
                if old is None:
                    super().__setattr__(name, value)
                elif old != value and value is not None:
                    raise argparse.ArgumentError(
                        None,
                        f"conflicting values provided for {arg.action.dest!r} ({old!r} and {value!r})",
                    )
            else:
                raise NotImplementedError(
                    f"Action {action_type!r} for {arg.action.dest!r} is not supported"
                )
        else:
            return super().__setattr__(name, value)


class DebputyArgumentParser(ColorizedArgumentParser):
    """
    Hacks around a standard ArgumentParser to allow to have a limited set of
    options both outside and inside subcommands

    Based on: https://www.enricozini.org/blog/2022/python/sharing-argparse-arguments-with-subcommands/
    """

    def __init__(self, *args: Any, **kw: Any) -> None:
        super().__init__(*args, **kw)

        if not hasattr(self, "shared_args"):
            self.shared_args: dict[str, SharedArgument] = {}

        # Add arguments from the shared ones
        for a in self.shared_args.values():
            super().add_argument(*a.args, **a.kwargs)

    def add_argument(self, *args: Any, **kw: Any) -> Any:
        shared = kw.pop("shared", False)
        res = super().add_argument(*args, **kw)
        if shared:
            action = kw.get("action")
            if action not in ("store", "store_true", "store_false", "append"):
                raise NotImplementedError(
                    f"Action {action!r} for {args!r} is not supported"
                )
            # Take note of the argument if it was marked as shared
            self.shared_args[res.dest] = SharedArgument(res, args, kw)
        return res

    def add_subparsers(self, *args: Any, **kw: Any) -> Any:
        if "parser_class" not in kw:
            kw["parser_class"] = type(
                "ArgumentParser",
                (self.__class__,),
                {"shared_args": dict(self.shared_args)},
            )
        return super().add_subparsers(*args, **kw)

    def parse_args(self, *args: Any, **kw: Any) -> Any:
        if "namespace" not in kw:
            # Use a subclass to pass the special action list without making it
            # appear as an argument
            kw["namespace"] = type(
                "Namespace", (Namespace,), {"_shared_args": self.shared_args}
            )()
        return super().parse_args(*args, **kw)


def _add_common_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--debputy-manifest",
        dest="debputy_manifest",
        action="store",
        default=None,
        help="Specify another `debputy` manifest (default: debian/debputy.manifest)",
        shared=True,
    )

    parser.add_argument(
        "-d",
        "--debug",
        dest="debug_mode",
        action="store_true",
        default=False,
        help="Enable debug logging and raw stack traces on errors. Some warnings become errors as a consequence.",
        shared=True,
    )

    parser.add_argument(
        "--no-pager",
        dest="pager",
        action="store_false",
        default=True,
        help="For subcommands that can use a pager, disable the use of pager. Some output formats implies --no-pager",
        shared=True,
    )

    parser.add_argument(
        "--plugin",
        dest="required_plugins",
        action="append",
        type=str,
        default=[],
        help="Request the plugin to be loaded. Can be used multiple time."
        " Ignored for some commands (such as autopkgtest-test-runner)",
        shared=True,
    )


def _add_packages_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-p",
        "--package",
        dest="packages",
        action="append",
        type=str,
        default=[],
        help="The package(s) to act on.  Affects default permission normalization rules",
    )


def _build_subcommand_log_level(_context: CommandContext) -> int:
    log_level: int = PRINT_BUILD_SYSTEM_COMMAND
    if os.environ.get("DH_VERBOSE", "") != "":
        log_level = PRINT_COMMAND
    return log_level


internal_commands = ROOT_COMMAND.add_dispatching_subcommand(
    "internal-command",
    dest="internal_command",
    metavar="command",
    help_description="Commands used for internal purposes. These are implementation details and subject to change",
)
tool_support_commands = ROOT_COMMAND.add_dispatching_subcommand(
    "tool-support",
    help_description="Tool integration commands. These are intended to have stable output and behavior",
    dest="tool_subcommand",
    metavar="command",
)


def parse_args() -> argparse.Namespace:
    description = textwrap.dedent(
        """\
    The `debputy` program is a Debian packaging tool.

    It serves multiple roles in the Debian packaging stack:

      1) It is a maintainer support tool that can help maintain packages by providing
         editor support (LSP), batch linting and style/file formatting for Debian
         packaging files. This is covers subcommands like `debputy lint`,
         `debputy lsp server`, `debputy reformat`

      2) It is a manifest-based Debian package builder aiming to replace existing
         package helper tools such as `debhelper . In this role, `debputy` is used
         as a part of compiling a source package and transforming it into one
         or more binary (.deb) packages.

    If you are using a screen reader, consider exporting setting the environment variable
    OPTIMIZE_FOR_SCREEN_READER=1. This will remove some of the visual formatting and some
    commands will render the output in a purely textual manner rather than visual layout.
    """
    )

    eplilog = textwrap.dedent(
        """\

    Bug tracker: https://salsa.debian.org/debian/debputy/-/issues
    """
    )

    parser: argparse.ArgumentParser = DebputyArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        allow_abbrev=False,
        epilog=eplilog,
        prog=program_name(),
    )

    parser.add_argument("--version", action="version", version=__version__)

    _add_common_args(parser)
    from debputy.commands.debputy_cmd.plugin_cmds import (
        ensure_plugin_commands_are_loaded,
    )
    from debputy.commands.debputy_cmd.lint_and_lsp_cmds import (
        ensure_lint_and_lsp_commands_are_loaded,
    )

    ensure_plugin_commands_are_loaded()
    ensure_lint_and_lsp_commands_are_loaded()

    ROOT_COMMAND.configure(parser)

    autocomplete(parser)

    argv = sys.argv
    try:
        i = argv.index("--")
        upstream_args = argv[i + 1 :]
        argv = argv[:i]
    except (IndexError, ValueError):
        upstream_args = []
    parsed_args: argparse.Namespace = parser.parse_args(argv[1:])

    setattr(parsed_args, "upstream_args", upstream_args)
    if hasattr(parsed_args, "packages"):
        setattr(parsed_args, "packages", frozenset(parsed_args.packages))

    if os.environ.get("DEBPUTY_DEBUG", "") == "1":
        setattr(parsed_args, "debug_mode", True)

    return parsed_args


@ROOT_COMMAND.register_subcommand(
    "check-manifest",
    help_description="Check the manifest for obvious errors, but do not run anything",
    requested_plugins_only=True,
)
def _check_manifest(context: CommandContext) -> None:
    context.parse_manifest()


def _install_plugin_from_plugin_metadata(
    plugin_metadata: DebputyPluginMetadata,
    dest_dir: str,
) -> None:
    related_files = find_related_implementation_files_for_plugin(plugin_metadata)
    install_dir = os.path.join(
        f"{dest_dir}/{DEBPUTY_PLUGIN_ROOT_DIR}".replace("//", "/"),
        "debputy",
        "plugins",
    )

    os.umask(0o022)
    os.makedirs(install_dir, exist_ok=True)
    cmd = ["cp", "--reflink=auto", "-t", install_dir]
    cmd.extend(related_files)
    cmd.append(plugin_metadata.plugin_path)
    _info(f"   {escape_shell(*cmd)}")
    subprocess.check_call(
        cmd,
        stdin=subprocess.DEVNULL,
    )


@internal_commands.register_subcommand(
    "install-plugin",
    help_description="[Internal command] Install a plugin and related files",
    requested_plugins_only=True,
    argparser=[
        add_arg("target_plugin", metavar="PLUGIN", action="store"),
        add_arg(
            "--dest-dir",
            dest="dest_dir",
            default="",
            action="store",
        ),
    ],
)
def _install_plugin(context: CommandContext) -> None:
    target_plugin = context.parsed_args.target_plugin
    if not os.path.isfile(target_plugin):
        _error(
            f'The value "{target_plugin}" must be a file. It should be the JSON descriptor of'
            f" the plugin."
        )
    plugin_metadata = parse_json_plugin_desc(target_plugin)
    _install_plugin_from_plugin_metadata(
        plugin_metadata,
        context.parsed_args.dest_dir,
    )


_DH_PLUGIN_PKG_DIR = "debputy-plugins"


def _find_plugins_and_tests_in_source_package(
    context: CommandContext,
) -> Tuple[bool, List[Tuple[DebputyPluginMetadata, str]], List[str]]:
    debian_dir = context.debian_dir
    binary_packages = context.binary_packages()
    installs = []
    all_tests = []
    had_plugin_dir = False
    for binary_package in binary_packages.values():
        if not binary_package.should_be_acted_on:
            continue
        debputy_plugins_dir = dhe_pkgdir(debian_dir, binary_package, _DH_PLUGIN_PKG_DIR)
        if debputy_plugins_dir is None:
            continue
        if not debputy_plugins_dir.is_dir:
            continue
        had_plugin_dir = True
        dest_dir = os.path.join("debian", binary_package.name)
        for path in debputy_plugins_dir.iterdir:
            if not path.is_file or not path.name.endswith((".json", ".json.in")):
                continue
            plugin_metadata = parse_json_plugin_desc(path.path)
            if (
                plugin_metadata.plugin_name.startswith("debputy-")
                or plugin_metadata.plugin_name == "debputy"
            ):
                _error(
                    f"The plugin name {plugin_metadata.plugin_name} is reserved by debputy. Please rename"
                    " the plugin to something else."
                )
            installs.append((plugin_metadata, dest_dir))
            all_tests.extend(find_tests_for_plugin(plugin_metadata))
    return had_plugin_dir, installs, all_tests


@ROOT_COMMAND.register_subcommand(
    "autopkgtest-test-runner",
    requested_plugins_only=True,
    help_description="Detect tests in the debian dir and run them against installed plugins",
)
def _autodep8_test_runner(context: CommandContext) -> None:
    ad_hoc_run = "AUTOPKGTEST_TMP" not in os.environ
    _a, _b, all_tests = _find_plugins_and_tests_in_source_package(context)

    source_package = context.source_package()
    explicit_test = (
        "autopkgtest-pkg-debputy" in source_package.fields.get("Testsuite", "").split()
    )

    if not shutil.which("py.test"):
        if ad_hoc_run:
            extra_context = ""
            if not explicit_test:
                extra_context = (
                    " Remember to add python3-pytest to the Depends field of your autopkgtests field if"
                    " you are writing your own test case for autopkgtest. Note you can also add"
                    ' "autopkgtest-pkg-debputy" to the "Testsuite" field in debian/control if you'
                    " want the test case autogenerated."
                )
            _error(
                f"Please install the py.test command (apt-get install python3-pytest).{extra_context}"
            )
        _error("Please add python3-pytest to the Depends field of your autopkgtests.")

    if not all_tests:
        extra_context = ""
        if explicit_test:
            extra_context = (
                " If the package no longer provides any plugin or tests, please remove the "
                ' "autopkgtest-pkg-debputy" test from the "Testsuite" in debian/control'
            )
        _error(
            "There are no tests to be run. The autodep8 feature should not have generated a test for"
            f" this case.{extra_context}"
        )

    if _run_tests(
        context,
        all_tests,
        test_plugin_location="installed",
        on_error_return=False,
    ):
        return
    extra_context = ""
    if not ad_hoc_run:
        extra_context = (
            ' These tests can be run manually via the "debputy autopkgtest-test-runner" command without any'
            ' autopkgtest layering. To do so, install "dh-debputy python3-pytest" plus the packages'
            " being tested and relevant extra dependencies required for the tests. Then open a shell in"
            f' the unpacked source directory of {source_package.name} and run "debputy autopkgtest-test-runner"'
        )
    _error(f"The tests were not successful.{extra_context}")


@internal_commands.register_subcommand(
    "dh-integration-install-plugin",
    help_description="[Internal command] Install a plugin and related files via debhelper integration",
    requested_plugins_only=True,
    argparser=_add_packages_args,
)
def _dh_integration_install_plugin(context: CommandContext) -> None:
    had_plugin_dir, installs, all_tests = _find_plugins_and_tests_in_source_package(
        context
    )

    if not installs:
        if had_plugin_dir:
            _warn(
                "There were plugin dirs, but no plugins were detected inside them. Please ensure that "
                f" the plugin dirs (debian/<pkg>.{_DH_PLUGIN_PKG_DIR} or debian/{_DH_PLUGIN_PKG_DIR})"
                f" contains a .json or .json.in file, or remove them (plus drop the"
                f" dh-sequence-installdebputy build dependency) if they are no longer useful."
            )
        else:
            _info(
                f"No plugin directories detected (debian/<pkg>.{_DH_PLUGIN_PKG_DIR} or debian/{_DH_PLUGIN_PKG_DIR})"
            )
        return

    if all_tests:
        if "nocheck" in context.deb_build_options_and_profiles.deb_build_options:
            _info("Skipping tests due to DEB_BUILD_OPTIONS=nocheck")
        elif not shutil.which("py.test"):
            _warn("Skipping tests because py.test is not available")
        else:
            _run_tests(context, all_tests)
    else:
        _info("No tests detected for any of the plugins. Skipping running tests.")

    for plugin_metadata, dest_dir in installs:
        _info(f"Installing plugin {plugin_metadata.plugin_name} into {dest_dir}")
        _install_plugin_from_plugin_metadata(plugin_metadata, dest_dir)


def _run_tests(
    context: CommandContext,
    test_paths: List[str],
    *,
    cwd: Optional[str] = None,
    tmpdir_root: Optional[str] = None,
    test_plugin_location: Literal["installed", "uninstalled"] = "uninstalled",
    on_error_return: Optional[Any] = None,
    on_success_return: Optional[Any] = True,
) -> Any:
    env = dict(os.environ)
    env["DEBPUTY_TEST_PLUGIN_LOCATION"] = test_plugin_location
    if "PYTHONPATH" in env:
        env["PYTHONPATH"] = f"{DEBPUTY_ROOT_DIR}:{env['PYTHONPATH']}"
    else:
        env["PYTHONPATH"] = str(DEBPUTY_ROOT_DIR)

    env["PYTHONDONTWRITEBYTECODE"] = "1"
    _info("Running debputy plugin tests.")
    _info("")
    _info("Environment settings:")
    for envname in [
        "PYTHONPATH",
        "PYTHONDONTWRITEBYTECODE",
        "DEBPUTY_TEST_PLUGIN_LOCATION",
    ]:
        _info(f"    {envname}={env[envname]}")

    with TemporaryDirectory(dir=tmpdir_root) as tmpdir:
        cmd = [
            "py.test",
            "-vvvvv" if context.parsed_args.debug_mode else "-v",
            "--config-file=/dev/null",
            f"--rootdir={cwd if cwd is not None else '.'}",
            "-o",
            f"cache_dir={tmpdir}",
        ]
        cmd.extend(test_paths)

        _info(f"Test Command: {escape_shell(*cmd)}")
        try:
            subprocess.check_call(
                cmd,
                stdin=subprocess.DEVNULL,
                env=env,
                cwd=cwd,
            )
        except subprocess.CalledProcessError:
            if on_error_return is None:
                _error("The tests were not successful.")
            return on_error_return
    return True


@internal_commands.register_subcommand(
    "run-tests-for-plugin",
    help_description="[Internal command] Run tests for a plugin",
    requested_plugins_only=True,
    argparser=[
        add_arg("target_plugin", metavar="PLUGIN", action="store"),
        add_arg(
            "--require-tests",
            dest="require_tests",
            default=True,
            action=argparse.BooleanOptionalAction,
        ),
    ],
)
def _run_tests_for_plugin(context: CommandContext) -> None:
    target_plugin = context.parsed_args.target_plugin
    if not os.path.isfile(target_plugin):
        _error(
            f'The value "{target_plugin}" must be a file. It should be the JSON descriptor of'
            f" the plugin."
        )
    try:
        plugin_metadata = find_json_plugin(
            context.plugin_search_dirs,
            target_plugin,
        )
    except PluginNotFoundError as e:
        _error(e.message)

    tests = find_tests_for_plugin(plugin_metadata)

    if not tests:
        if context.parsed_args.require_tests:
            plugin_name = plugin_metadata.plugin_name
            plugin_dir = os.path.dirname(plugin_metadata.plugin_path)

            _error(
                f"Cannot find any tests for {plugin_name}: Expected them to be in "
                f' "{plugin_dir}". Use --no-require-tests to consider missing tests'
                " a non-error."
            )
        _info(
            f"No tests found for {plugin_metadata.plugin_name}. Use --require-tests to turn"
            " this into an error."
        )
        return

    if not shutil.which("py.test"):
        _error(
            f"Cannot run the tests for {plugin_metadata.plugin_name}: This feature requires py.test"
            f" (apt-get install python3-pytest)"
        )
    _run_tests(context, tests, cwd="/")


@internal_commands.register_subcommand(
    "process-template",
    help_description="[Internal command] Process a debputy template",
    requested_plugins_only=True,
    argparser=[
        add_arg(
            "template_file",
            metavar="template-file",
            help="The source template to be processed.",
        ),
        add_arg(
            "output_file",
            metavar="output-file",
            help="Where to place the resulting file should be written",
        ),
    ],
)
def _process_template(context: CommandContext) -> None:
    parsed_args = context.parsed_args
    template_file = parsed_args.template_file
    output_file = parsed_args.output_file

    try:
        from jinja2 import Environment, FileSystemLoader, Template
    except ImportError:
        _error("This sub-command requires a compatible version of python3-jinja2")

    feature_set = context.load_plugins()

    with open(template_file) as fd:
        jinja_template = Template(
            fd.read(),
            autoescape=False,
            block_start_string="<%",
            block_end_string="%>",
            variable_start_string="<<",
            variable_end_string=">>",
            comment_start_string="<#",
            comment_end_string="#>",
        )

    def _render_pmr(rule_name: str, heading: str) -> str:
        pmr_rule = lookup_pmr_rule(feature_set, rule_name)
        return render_rule(
            pmr_rule.rule_name,
            pmr_rule.parser,
            pmr_rule.plugin_metadata,
            is_root_rule=pmr_rule.is_root_rule,
            heading=heading,
            include_ref_doc_link=False,
            include_alt_format=False,
            manifest_format_url="",
        )

    with open(output_file, "w") as fd:
        fd.write(jinja_template.render(render_pmr=_render_pmr))


@internal_commands.register_subcommand(
    "dpkg-build-driver-run-task",
    help_description="[Internal command] Perform a given Dpkg::BuildDriver task (Not stable API)",
    requested_plugins_only=True,
    default_log_level=_build_subcommand_log_level,
    argparser=[
        add_arg(
            "task_name",
            metavar="task-name",
            choices=[
                "clean",
                "build",
                "build-arch",
                "build-indep",
                "binary",
                "binary-arch",
                "binary-indep",
            ],
            help="The task to run",
        ),
        add_arg(
            "output",
            nargs="?",
            default="..",
            metavar="output",
            help="Where to place the resulting packages. Should be a directory",
        ),
    ],
)
def _dpkg_build_driver_integration(context: CommandContext) -> None:
    parsed_args = context.parsed_args
    log_level = context.set_log_level_for_build_subcommand()
    task_name = parsed_args.task_name

    if task_name.endswith("-indep"):
        context.package_set = "indep"
    elif task_name.endswith("arch"):
        context.package_set = "arch"

    manifest = context.parse_manifest()

    plugins = context.load_plugins().plugin_data
    for plugin in plugins.values():
        if not plugin.is_bundled:
            _info(f"Loaded plugin {plugin.plugin_name}")
    if task_name == "clean":
        perform_clean(context, manifest)
    elif task_name in ("build", "build-indep", "build-arch"):
        perform_builds(context, manifest)
    elif task_name in ("binary", "binary-indep", "binary-arch"):
        perform_builds(context, manifest)
        assemble(
            context,
            manifest,
            INTEGRATION_MODE_FULL,
            debug_materialization=log_level is not None,
        )
    else:
        _error(f"Unsupported Dpkg::BuildDriver task: {task_name}.")


@internal_commands.register_subcommand(
    "dh-integration-generate-debs",
    help_description="[Internal command] Generate .deb/.udebs packages from debian/<pkg> (Not stable API)",
    requested_plugins_only=True,
    default_log_level=_build_subcommand_log_level,
    argparser=[
        _add_packages_args,
        add_arg(
            "output",
            metavar="output",
            help="Where to place the resulting packages. Should be a directory",
        ),
        # Added for "help only" - you cannot trigger this option in practice
        add_arg(
            "--",
            metavar="UPSTREAM_ARGS",
            action="extend",
            nargs="+",
            dest="unused",
        ),
    ],
)
def _dh_integration_generate_debs(context: CommandContext) -> None:
    integrated_with_debhelper()
    log_level = context.set_log_level_for_build_subcommand()
    integration_mode = context.resolve_integration_mode()
    is_dh_rrr_only_mode = integration_mode == INTEGRATION_MODE_DH_DEBPUTY_RRR
    if is_dh_rrr_only_mode:
        problematic_plugins = list(context.requested_plugins())
        problematic_plugins.extend(context.required_plugins())
        if problematic_plugins:
            plugin_names = ", ".join(problematic_plugins)
            _error(
                f"Plugins are not supported in the zz-debputy-rrr sequence. Detected plugins: {plugin_names}"
            )

    plugins = context.load_plugins().plugin_data
    for plugin in plugins.values():
        if not plugin.is_bundled:
            _info(f"Loaded plugin {plugin.plugin_name}")
    manifest = context.parse_manifest()

    assemble(
        context,
        manifest,
        integration_mode,
        debug_materialization=log_level is not None,
    )


def assemble(
    context: CommandContext,
    manifest: HighLevelManifest,
    integration_mode: DebputyIntegrationMode,
    *,
    debug_materialization: bool = False,
) -> None:
    source_fs = FSROOverlay.create_root_dir("..", ".")
    source_version = manifest.source_version()
    is_native = "-" not in source_version
    is_dh_rrr_only_mode = integration_mode == INTEGRATION_MODE_DH_DEBPUTY_RRR
    package_data_table = manifest.perform_installations(integration_mode)
    if not is_dh_rrr_only_mode:
        for dctrl_bin in manifest.active_packages:
            package = dctrl_bin.name
            dctrl_data = package_data_table[package]
            fs_root = dctrl_data.fs_root
            package_metadata_context = dctrl_data.package_metadata_context

            assert dctrl_bin.should_be_acted_on

            detect_systemd_user_service_files(dctrl_bin, fs_root)
            usr_local_transformation(dctrl_bin, fs_root)
            handle_perl_code(
                dctrl_bin,
                manifest.dpkg_architecture_variables,
                fs_root,
                dctrl_data.substvars,
            )
            if "nostrip" not in manifest.deb_options_and_profiles.deb_build_options:
                dbgsym_ids = relocate_dwarves_into_dbgsym_packages(
                    dctrl_bin,
                    fs_root,
                    dctrl_data.dbgsym_info.dbgsym_fs_root,
                )
                dctrl_data.dbgsym_info.dbgsym_ids = dbgsym_ids

            fixup_debian_changelog_and_news_file(
                dctrl_bin,
                fs_root,
                is_native,
                manifest.deb_options_and_profiles,
            )
            if not is_native:
                install_upstream_changelog(
                    dctrl_bin,
                    fs_root,
                    source_fs,
                )
            run_package_processors(manifest, package_metadata_context, fs_root)

        cross_package_control_files(package_data_table, manifest)
    for binary_data in package_data_table:
        if not binary_data.binary_package.should_be_acted_on:
            continue
        # Ensure all fs's are read-only before we enable cross package checks.
        # This ensures that no metadata detector will never see a read-write FS
        pkg_fs_root: "FSRootDir" = cast("FSRootDir", binary_data.fs_root)
        pkg_fs_root.is_read_write = False

    package_data_table.enable_cross_package_checks = True
    assemble_debs(
        context,
        manifest,
        package_data_table,
        is_dh_rrr_only_mode,
        debug_materialization=debug_materialization,
    )


@tool_support_commands.register_subcommand(
    "supports-tool-command",
    help_description="Test where a given tool-support command exists",
    argparser=add_arg(
        "test_command",
        metavar="name",
        default=None,
        help="The name of the command",
    ),
)
def _supports_tool_command(context: CommandContext) -> None:
    command_name = context.parsed_args.test_command
    if tool_support_commands.has_command(command_name):
        sys.exit(0)
    else:
        sys.exit(2)


@tool_support_commands.register_subcommand(
    "export-reference-data",
    help_description="Export reference data for other tool-support commands",
    argparser=[
        add_arg(
            "--output-format",
            default="text",
            choices=["text", "json"],
            help="Output format of the reference data",
        ),
        add_arg(
            "dataset",
            metavar="name",
            default=None,
            nargs="?",
            help="The dataset to export (if any)",
            choices=REFERENCE_DATA_TABLE,
        ),
    ],
)
def _export_reference_data(context: CommandContext) -> None:
    dataset_name = context.parsed_args.dataset
    output_format = context.parsed_args.output_format
    if dataset_name is not None:
        subdata_set = REFERENCE_DATA_TABLE.get(dataset_name)
        if subdata_set is None:
            _error(f"Unknown data set: {dataset_name}")
        reference_data = {
            dataset_name: subdata_set,
        }
    else:
        subdata_set = None
        reference_data = REFERENCE_DATA_TABLE
    if output_format == "text":
        if subdata_set is None:
            _error(
                "When output format is text, then the dataset name is required (it is optional for JSON formats)."
            )
        with _stream_to_pager(context.parsed_args) as (fd, fo):
            header = ["key", "description"]
            rows = [(k, v["description"]) for k, v in subdata_set.items()]
            fo.print_list_table(header, rows)
            fo.print()
            fo.print("If you wanted this as JSON, please use --output-format=json")
    elif output_format == "json":
        _json_output(
            {
                "reference-data": reference_data,
            }
        )
    else:
        raise AssertionError(f"Unsupported output format {output_format}")


@tool_support_commands.register_subcommand(
    "annotate-debian-directory",
    log_only_to_stderr=True,
    help_description="Scan debian/* for known package files and annotate them with information."
    " Output is evaluated and may change. Please get in touch if you want to use it"
    " or want additional features.",
)
def _annotate_debian_directory(context: CommandContext) -> None:
    # Validates that we are run from a debian directory as a side effect
    binary_packages = context.binary_packages()
    feature_set = context.load_plugins()

    result = scan_debian_dir(
        feature_set,
        context.source_package(),
        binary_packages,
        context.debian_dir,
    )
    annotated, reference_data_set_names, dh_assistant_exit_code, dh_issues = result

    data = {
        "result": annotated,
        "reference-datasets": reference_data_set_names,
    }
    if dh_issues is not None or dh_assistant_exit_code != 0:
        data["issues"] = [
            {
                "source": "dh_assistant",
                "exit-code": dh_assistant_exit_code,
                "issue-data": dh_issues,
            }
        ]
    _json_output(data)


def _json_output(data: Any) -> None:
    if sys.stdout.isatty():
        # sort_keys might be tempting but generally insert order makes more sense in practice.
        json.dump(data, sys.stdout, indent=4)
    else:
        json.dump(data, sys.stdout)
    if sys.stdout.isatty():
        # Looks better with a final newline.
        print()


@ROOT_COMMAND.register_subcommand(
    "migrate-from-dh",
    help_description='Generate/update manifest from a "dh $@" using package',
    argparser=[
        add_arg(
            "--acceptable-migration-issues",
            dest="acceptable_migration_issues",
            action="append",
            type=str,
            default=[],
            help="Continue the migration even if this/these issues are detected."
            " Can be set to ALL (in all upper-case) to accept all issues",
        ),
        add_arg(
            "--migration-target",
            dest="migration_target",
            action="store",
            choices=MIGRATORS,
            type=str,
            default=None,
            help="Continue the migration even if this/these issues are detected."
            " Can be set to ALL (in all upper-case) to accept all issues",
        ),
        add_arg(
            "--no-act",
            "--no-apply-changes",
            dest="destructive",
            action="store_false",
            default=None,
            help="Do not perform changes.  Existing manifest will not be overridden",
        ),
        add_arg(
            "--apply-changes",
            dest="destructive",
            action="store_true",
            default=None,
            help="Perform changes.  The debian/debputy.manifest will updated in place if exists",
        ),
    ],
)
def _migrate_from_dh(context: CommandContext) -> None:
    context.must_be_called_in_source_root()
    parsed_args = context.parsed_args
    resolved_migration_target = _check_migration_target(
        context,
        parsed_args.migration_target,
    )
    context.debputy_integration_mode = resolved_migration_target
    manifest = context.parse_manifest()
    acceptable_migration_issues = AcceptableMigrationIssues(
        frozenset(
            i for x in parsed_args.acceptable_migration_issues for i in x.split(",")
        )
    )
    migrate_from_dh(
        _output_styling(context.parsed_args, sys.stdout),
        manifest,
        acceptable_migration_issues,
        parsed_args.destructive,
        resolved_migration_target,
        lambda p: context.parse_manifest(manifest_path=p),
    )


def _setup_and_parse_args() -> argparse.Namespace:
    is_arg_completing = "_ARGCOMPLETE" in os.environ
    if not is_arg_completing:
        setup_logging()
    parsed_args = parse_args()
    if is_arg_completing:
        # We could be asserting at this point; but lets just recover gracefully.
        setup_logging()
    return parsed_args


def main() -> None:
    parsed_args = _setup_and_parse_args()
    plugin_search_dirs = [str(DEBPUTY_PLUGIN_ROOT_DIR)]
    try:
        cmd_arg = CommandArg(
            parsed_args,
            plugin_search_dirs,
        )
        ROOT_COMMAND(cmd_arg)
    except PluginInitializationError as e:
        _error_w_stack_trace(
            "Failed to load a plugin - relevant stack trace (use --debug/DEBPUTY_DEBUG=1 to see the full one):",
            e.message,
            e.__cause__ if e.__cause__ else e,
            parsed_args.debug_mode,
            follow_warning=[
                "Please consider filing a bug against the plugin in question"
            ],
        )
    except UnhandledOrUnexpectedErrorFromPluginError as e:
        trace = e.__cause__ if e.__cause__ is not None else e
        # TODO: Reframe this as an internal error if `debputy` is the misbehaving plugin
        if isinstance(trace, SymlinkLoopError):
            _error_w_stack_trace(
                "Error in `debputy`:",
                e.message,
                trace,
                parsed_args.debug_mode,
                orig_exception=e,
                follow_warning=[
                    "Please consider filing a bug against `debputy` in question"
                ],
            )
        else:
            _error_w_stack_trace(
                "A plugin misbehaved:",
                e.message,
                trace,
                parsed_args.debug_mode,
                orig_exception=e,
                follow_warning=[
                    "Please consider filing a bug against the plugin in question"
                ],
            )
    except PluginAPIViolationError as e:
        trace = e.__cause__ if e.__cause__ is not None else e
        # TODO: Reframe this as an internal error if `debputy` is the misbehaving plugin
        _error_w_stack_trace(
            "A plugin misbehaved:",
            e.message,
            trace,
            parsed_args.debug_mode,
            orig_exception=e,
            follow_warning=[
                "Please consider filing a bug against the plugin in question"
            ],
        )
    except DebputyRuntimeError as e:
        if parsed_args.debug_mode:
            _warn(
                "Re-raising original exception to show the full stack trace due to debug mode being active"
            )
            raise e
        if isinstance(e, DebputyRuntimeErrorWithPreamble):
            e.render_preamble()
        _error(e.message)
    except AssertionError as e:
        _error_w_stack_trace(
            "Internal error in debputy",
            str(e),
            e,
            parsed_args.debug_mode,
            orig_exception=e,
            follow_warning=["Please file a bug against debputy with the full output."],
        )
    except subprocess.CalledProcessError as e:
        cmd = escape_shell(*e.cmd) if isinstance(e.cmd, list) else str(e.cmd)
        _error_w_stack_trace(
            f"The command << {cmd} >> failed and the code did not explicitly handle that exception.",
            str(e),
            e,
            parsed_args.debug_mode,
            orig_exception=e,
            follow_warning=[
                "The output above this error and the stacktrace may provide context to why the command failed.",
                "Please file a bug against debputy with the full output.",
            ],
        )
    except Exception as e:
        _error_w_stack_trace(
            "Unhandled exception (Re-run with --debug/DEBPUTY_DEBUG=1 to see the raw stack trace)",
            str(e),
            e,
            parsed_args.debug_mode,
            orig_exception=e,
            follow_warning=["Please file a bug against debputy with the full output."],
        )


def _error_w_stack_trace(
    warning: str,
    error_msg: str,
    stacktrace: BaseException,
    debug_mode: bool,
    orig_exception: Optional[BaseException] = None,
    follow_warning: Optional[List[str]] = None,
) -> "NoReturn":
    if debug_mode:
        _warn(
            "Re-raising original exception to show the full stack trace due to debug mode being active"
        )
        raise orig_exception if orig_exception is not None else stacktrace
    _warn(warning)
    _warn("  ----- 8< ---- BEGIN STACK TRACE ---- 8< -----")
    traceback.print_exception(stacktrace)
    _warn("  ----- 8< ---- END STACK TRACE ---- 8< -----")
    if follow_warning:
        for line in follow_warning:
            _warn(line)
    _error(error_msg)


if __name__ == "__main__":
    main()



import argparse
import collections
import functools
import glob
import logging
import os
import re
import shutil
import subprocess
import sys
import time
from itertools import zip_longest
from pathlib import Path
from typing import (
    NoReturn,
    TYPE_CHECKING,
    Union,
    Set,
    FrozenSet,
    Optional,
    TypeVar,
    Dict,
    Iterator,
    Iterable,
    Literal,
    Tuple,
    Sequence,
    List,
    Mapping,
    Any,
)

from debian.deb822 import Deb822

from debputy import DEBPUTY_DOC_ROOT_DIR
from debputy.architecture_support import DpkgArchitectureBuildProcessValuesTable
from debputy.exceptions import DebputySubstitutionError

try:
    from Levenshtein import distance
except ImportError:

    CAN_DETECT_TYPOS = False

    def detect_possible_typo(
        provided_value: str,
        known_values: Iterable[str],
        *,
        max_edit_distance: int = 2,
    ) -> Sequence[str]:
        return ()

else:

    CAN_DETECT_TYPOS = True

    def detect_possible_typo(
        provided_value: str,
        known_values: Iterable[str],
        *,
        max_edit_distance: int = 2,
    ) -> Sequence[str]:
        k_len = len(provided_value)
        candidates = []
        for known_value in known_values:
            if abs(k_len - len(known_value)) > max_edit_distance:
                continue
            d = distance(provided_value, known_value)
            if d > max_edit_distance:
                continue
            candidates.append(known_value)
        return candidates


if TYPE_CHECKING:
    from debputy.types import EnvironmentModification
    from debputy.packages import BinaryPackage
    from debputy.substitution import Substitution


T = TypeVar("T")


SLASH_PRUNE = re.compile("//+")
PKGNAME_REGEX = re.compile(r"[a-z0-9][-+.a-z0-9]+", re.ASCII)
PKGVERSION_REGEX = re.compile(
    r"""
                 (?: \d+ : )?                # Optional epoch
                 \d[0-9A-Za-z.+:~]*          # Upstream version (with no hyphens)
                 (?: - [0-9A-Za-z.+:~]+ )*   # Optional debian revision (+ upstreams versions with hyphens)
""",
    re.VERBOSE | re.ASCII,
)
DEFAULT_PACKAGE_TYPE = "deb"
DBGSYM_PACKAGE_TYPE = "deb"
UDEB_PACKAGE_TYPE = "udeb"

POSTINST_DEFAULT_CONDITION = (
    '[ "$1" = "configure" ]'
    ' || [ "$1" = "abort-upgrade" ]'
    ' || [ "$1" = "abort-deconfigure" ]'
    ' || [ "$1" = "abort-remove" ]'
)


_SPACE_RE = re.compile(r"\s")
_WORD_EQUAL = re.compile(r"^-*[\w_\-]+=")
_DOUBLE_ESCAPEES = re.compile(r'([\n`$"\\])')
_REGULAR_ESCAPEES = re.compile(r"""([\s!"$()*+#;<>?@'\[\]\\`|~])""")
_PROFILE_GROUP_SPLIT = re.compile(r">\s+<")
_DEFAULT_LOGGER: Optional[logging.Logger] = None
_STDOUT_HANDLER: Optional[logging.StreamHandler[Any]] = None
_STDERR_HANDLER: Optional[logging.StreamHandler[Any]] = None
PRINT_COMMAND = logging.INFO + 3
PRINT_BUILD_SYSTEM_COMMAND = PRINT_COMMAND + 3
TRACE_LOG = logging.DEBUG + 3

# Map them back to `INFO`. The names must be unique so the prefix is stripped.
logging.addLevelName(PRINT_COMMAND, "__INFO")
logging.addLevelName(PRINT_BUILD_SYSTEM_COMMAND, "_INFO")
logging.addLevelName(TRACE_LOG, "TRACE")


def assume_not_none(x: Optional[T]) -> T:
    if x is None:  # pragma: no cover
        raise ValueError(
            'Internal error: None was given, but the receiver assumed "not None" here'
        )
    return x


def _non_verbose_info(msg: str) -> None:
    global _DEFAULT_LOGGER
    logger = _DEFAULT_LOGGER
    if logger is not None:
        logger.log(PRINT_BUILD_SYSTEM_COMMAND, msg)


def _info(msg: str) -> None:
    global _DEFAULT_LOGGER
    logger = _DEFAULT_LOGGER
    if logger:
        logger.info(msg)
    # No fallback print for info


def _is_trace_log_enabled() -> bool:
    global _DEFAULT_LOGGER
    logger = _DEFAULT_LOGGER
    return logger is not None and logger.isEnabledFor(TRACE_LOG)


def _trace_log(msg: str) -> None:
    global _DEFAULT_LOGGER
    logger = _DEFAULT_LOGGER
    if logger:
        logger.log(TRACE_LOG, msg)
    # No fallback print for this level


def _is_debug_log_enabled() -> bool:
    global _DEFAULT_LOGGER
    logger = _DEFAULT_LOGGER
    return logger is not None and logger.isEnabledFor(logging.DEBUG)


def _debug_log(msg: str) -> None:
    global _DEFAULT_LOGGER
    logger = _DEFAULT_LOGGER
    if logger:
        logger.debug(msg)
    # No fallback print for this level


def _error(msg: str, *, prog: Optional[str] = None) -> "NoReturn":
    global _DEFAULT_LOGGER
    logger = _DEFAULT_LOGGER
    if logger:
        logger.error(msg)
    else:
        me = os.path.basename(sys.argv[0]) if prog is None else prog
        print(
            f"{me}: error: {msg}",
            file=sys.stderr,
        )
    sys.exit(1)


def _warn(msg: str, *, prog: Optional[str] = None) -> None:
    global _DEFAULT_LOGGER
    logger = _DEFAULT_LOGGER
    if logger:
        logger.warning(msg)
    else:
        me = os.path.basename(sys.argv[0]) if prog is None else prog

        print(
            f"{me}: warning: {msg}",
            file=sys.stderr,
        )


class ColorizedArgumentParser(argparse.ArgumentParser):
    def error(self, message: str) -> NoReturn:
        self.print_usage(sys.stderr)
        _error(message, prog=self.prog)


def ensure_dir(path: str) -> None:
    if not os.path.isdir(path):
        os.makedirs(path, mode=0o755, exist_ok=True)


def _clean_path(orig_p: str) -> str:
    p = SLASH_PRUNE.sub("/", orig_p)
    if "." in p:
        path_base = p
        # We permit a single leading "./" because we add that when we normalize a path, and we want normalization
        # of a normalized path to be a no-op.
        if path_base.startswith("./"):
            path_base = path_base[2:]
            assert path_base
        for segment in path_base.split("/"):
            if segment in (".", ".."):
                raise ValueError(
                    'Please provide paths that are normalized (i.e., no ".." or ".").'
                    f' Offending input "{orig_p}"'
                )
    return p


def _normalize_path(path: str, with_prefix: bool = True) -> str:
    path = path.strip("/")
    if not path or path == ".":
        return "."
    if "//" in path or "." in path:
        path = _clean_path(path)
    if with_prefix ^ path.startswith("./"):
        if with_prefix:
            path = "./" + path
        else:
            path = path[2:]
    return path


def _normalize_link_target(link_target: str) -> str:
    link_target = SLASH_PRUNE.sub("/", link_target.lstrip("/"))
    result: List[str] = []
    for segment in link_target.split("/"):
        if segment in (".", ""):
            # Ignore these - the empty string is generally a trailing slash
            continue
        if segment == "..":
            # We ignore "root escape attempts" like the OS would (mapping /.. -> /)
            if result:
                result.pop()
        else:
            result.append(segment)
    return "/".join(result)


def manifest_format_doc(anchor: str) -> str:
    manifest_format = f"{DEBPUTY_DOC_ROOT_DIR}/MANIFEST-FORMAT.md"
    return f"{manifest_format}#{anchor}" if anchor else manifest_format


def _backslash_escape(m: re.Match[str]) -> str:
    return "\\" + m.group(0)


def _escape_shell_word(w: str) -> str:
    if _SPACE_RE.search(w):
        if "=" in w and (m := _WORD_EQUAL.search(w)) is not None:
            s = m.span(0)
            assert s[0] == 0
            prefix = w[0 : s[1]]
            escaped_value = _DOUBLE_ESCAPEES.sub(_backslash_escape, w[s[1] :])
            return f'{prefix}"{escaped_value}"'
        w = _DOUBLE_ESCAPEES.sub(_backslash_escape, w)
        return f'"{w}"'
    return _REGULAR_ESCAPEES.sub(_backslash_escape, w)


def escape_shell(*args: str) -> str:
    return " ".join(_escape_shell_word(w) for w in args)


def render_command(
    *args: str,
    cwd: Optional[str] = None,
    env_mod: Optional["EnvironmentModification"] = None,
) -> str:
    env_mod_prefix = ""
    if env_mod:
        env_mod_parts = []
        if bool(env_mod.removals):
            env_mod_parts.append("env")
            if cwd is not None:
                env_mod_parts.append(f"--chdir={escape_shell(cwd)}")
            env_mod_parts.extend(f"--unset={escape_shell(v)}" for v in env_mod.removals)
        env_mod_parts.extend(
            f"{escape_shell(k)}={escape_shell(v)}" for k, v in env_mod.replacements
        )

    chdir_prefix = ""
    if cwd is not None and cwd != ".":
        chdir_prefix = f"cd {escape_shell(cwd)} && "
    return f"{chdir_prefix}{env_mod_prefix}{escape_shell(*args)}"


def print_command(
    *args: str,
    cwd: Optional[str] = None,
    env_mod: Optional["EnvironmentModification"] = None,
    print_at_log_level: int = PRINT_COMMAND,
) -> None:
    if _DEFAULT_LOGGER is None or not _DEFAULT_LOGGER.isEnabledFor(print_at_log_level):
        return

    rendered_cmd = render_command(
        *args,
        cwd=cwd,
        env_mod=env_mod,
    )
    global _STDOUT_HANDLER
    handler = _STDOUT_HANDLER
    if handler is not None:
        handler.flush()
    # Ensure command is output immediately so it is hanging after its output.
    # TODO: This should `file` in case something in debputy redirects stdout
    # (nothing does that for now)
    print(f"   {rendered_cmd}")
    sys.stdout.flush()


def run_command(
    *args: str,
    cwd: Optional[str] = None,
    env: Optional[Mapping[str, str]] = None,
    env_mod: Optional["EnvironmentModification"] = None,
    print_at_log_level: int = PRINT_COMMAND,
    raise_file_not_found_on_missing_command: bool = False,
) -> None:
    print_command(
        *args,
        cwd=cwd,
        env_mod=env_mod,
        print_at_log_level=print_at_log_level,
    )
    if env_mod:
        if env is None:
            env = os.environ
        env = env_mod.compute_env(env)
        if env is os.environ:
            env = None
    try:
        subprocess.check_call(args, cwd=cwd, env=env)
    # At least "clean_logic.py" relies on catching FileNotFoundError
    except KeyboardInterrupt:
        _error(f"Interrupted (SIGINT) while running {escape_shell(*args)}")
    except FileNotFoundError:
        if raise_file_not_found_on_missing_command:
            raise
        if "/" in args[0]:
            _error(f"Could not run {escape_shell(args[0])}: Path does not exist")
        # Use the `command not found` to aid existing log pattern
        _error(f"{escape_shell(args[0])}: command not found")
    except subprocess.CalledProcessError as e:
        _error(f"The command {escape_shell(*args)} failed with status: {e.returncode}")


def run_build_system_command(
    *args: str,
    cwd: Optional[str] = None,
    env: Optional[Mapping[str, str]] = None,
    env_mod: Optional["EnvironmentModification"] = None,
    print_at_log_level: int = PRINT_BUILD_SYSTEM_COMMAND,
    raise_file_not_found_on_missing_command: bool = False,
) -> None:
    run_command(
        *args,
        cwd=cwd,
        env=env,
        env_mod=env_mod,
        print_at_log_level=print_at_log_level,
        raise_file_not_found_on_missing_command=raise_file_not_found_on_missing_command,
    )


def debian_policy_normalize_symlink_target(
    link_path: str,
    link_target: str,
    normalize_link_path: bool = False,
) -> str:
    if normalize_link_path:
        link_path = _normalize_path(link_path)
    elif not link_path.startswith("./"):
        raise ValueError("Link part was not normalized")

    link_path = link_path[2:]

    if not link_target.startswith("/"):
        link_target = "/" + os.path.dirname(link_path) + "/" + link_target

    link_path_parts = link_path.split("/")
    link_target_parts = [
        s for s in _normalize_link_target(link_target).split("/") if s != "."
    ]

    assert link_path_parts

    if link_target_parts and link_path_parts[0] == link_target_parts[0]:
        # Per Debian Policy, must be relative

        # First determine the length of the overlap
        common_segment_count = 1
        shortest_path_length = min(len(link_target_parts), len(link_path_parts))
        while (
            common_segment_count < shortest_path_length
            and link_target_parts[common_segment_count]
            == link_path_parts[common_segment_count]
        ):
            common_segment_count += 1

        if common_segment_count == shortest_path_length and len(
            link_path_parts
        ) - 1 == len(link_target_parts):
            normalized_link_target = "."
        else:
            up_dir_count = len(link_path_parts) - 1 - common_segment_count
            normalized_link_target_parts = []
            if up_dir_count:
                up_dir_part = "../" * up_dir_count
                # We overshoot with a single '/', so rstrip it away
                normalized_link_target_parts.append(up_dir_part.rstrip("/"))
            # Add the relevant down parts
            normalized_link_target_parts.extend(
                link_target_parts[common_segment_count:]
            )

            normalized_link_target = "/".join(normalized_link_target_parts)
    else:
        # Per Debian Policy, must be absolute
        normalized_link_target = "/" + "/".join(link_target_parts)

    return normalized_link_target


def has_glob_magic(pattern: str) -> bool:
    return glob.has_magic(pattern) or "{" in pattern


def glob_escape(replacement_value: str) -> str:
    if not glob.has_magic(replacement_value) or "{" not in replacement_value:
        return replacement_value
    return (
        replacement_value.replace("[", "[[]")
        .replace("]", "[]]")
        .replace("*", "[*]")
        .replace("?", "[?]")
        .replace("{", "[{]")
        .replace("}", "[}]")
    )


# TODO: This logic should probably be moved to `python-debian`
def active_profiles_match(
    profiles_raw: str,
    active_build_profiles: Union[Set[str], FrozenSet[str]],
) -> bool:
    profiles_raw = profiles_raw.strip()
    if profiles_raw[0] != "<" or profiles_raw[-1] != ">" or profiles_raw == "<>":
        raise ValueError(
            'Invalid Build-Profiles: Must start start and end with "<" + ">" but cannot be a literal "<>"'
        )
    profile_groups = _PROFILE_GROUP_SPLIT.split(profiles_raw[1:-1])
    for profile_group_raw in profile_groups:
        should_process_package = True
        for profile_name in profile_group_raw.split():
            negation = False
            if profile_name[0] == "!":
                negation = True
                profile_name = profile_name[1:]

            matched_profile = profile_name in active_build_profiles
            if matched_profile == negation:
                should_process_package = False
                break

        if should_process_package:
            return True

    return False


def _parse_build_profiles(build_profiles_raw: str) -> FrozenSet[FrozenSet[str]]:
    profiles_raw = build_profiles_raw.strip()
    if profiles_raw[0] != "<" or profiles_raw[-1] != ">" or profiles_raw == "<>":
        raise ValueError(
            'Invalid Build-Profiles: Must start start and end with "<" + ">" but cannot be a literal "<>"'
        )
    profile_groups = _PROFILE_GROUP_SPLIT.split(profiles_raw[1:-1])
    return frozenset(frozenset(g.split()) for g in profile_groups)


def resolve_source_date_epoch(
    command_line_value: Optional[int],
    *,
    substitution: Optional["Substitution"] = None,
) -> int:
    mtime = command_line_value
    if mtime is None and "SOURCE_DATE_EPOCH" in os.environ:
        sde_raw = os.environ["SOURCE_DATE_EPOCH"]
        if sde_raw == "":
            _error("SOURCE_DATE_EPOCH is set but empty.")
        mtime = int(sde_raw)
    if mtime is None and substitution is not None:
        try:
            sde_raw = substitution.substitute(
                "{{SOURCE_DATE_EPOCH}}",
                "Internal resolution",
            )
            mtime = int(sde_raw)
        except (DebputySubstitutionError, ValueError):
            pass
    if mtime is None:
        mtime = int(time.time())
    os.environ["SOURCE_DATE_EPOCH"] = str(mtime)
    return mtime


def compute_output_filename(control_root_dir: str, is_udeb: bool) -> str:
    with open(os.path.join(control_root_dir, "control"), "rt") as fd:
        control_file = Deb822(fd)

    package_name = control_file["Package"]
    package_version = control_file["Version"]
    package_architecture = control_file["Architecture"]
    extension = control_file.get("Package-Type") or "deb"
    if ":" in package_version:
        package_version = package_version.split(":", 1)[1]
    if is_udeb:
        extension = "udeb"

    return f"{package_name}_{package_version}_{package_architecture}.{extension}"


_SCRATCH_DIR = None
_DH_INTEGRATION_MODE = False


def integrated_with_debhelper() -> None:
    global _DH_INTEGRATION_MODE
    _DH_INTEGRATION_MODE = True


def scratch_dir(*, create_if_not_exists: bool = True) -> str:
    global _SCRATCH_DIR
    if _SCRATCH_DIR is not None:
        return _SCRATCH_DIR
    debputy_scratch_dir = "debian/.debputy/scratch-dir"
    is_debputy_dir = True
    if os.path.isdir("debian/.debputy") and not _DH_INTEGRATION_MODE:
        _SCRATCH_DIR = debputy_scratch_dir
    elif os.path.isdir("debian/.debhelper") or _DH_INTEGRATION_MODE:
        _SCRATCH_DIR = "debian/.debhelper/_debputy/scratch-dir"
        is_debputy_dir = False
    else:
        _SCRATCH_DIR = debputy_scratch_dir
    if create_if_not_exists:
        ensure_dir(_SCRATCH_DIR)
        if is_debputy_dir:
            generated_root_directory("debian/.debputy", internal_only=True)
    return _SCRATCH_DIR


def generated_root_directory(path: str, *, internal_only: bool = False) -> None:
    root_dir = Path(path)
    (root_dir / ".gitignore").write_text("*\n")
    # TODO: Should we add a "CACHEDIR.TAG" here? (Requires a relevant ignore rule
    if internal_only:
        (root_dir / "CACHEDIR.TAG").write_bytes(
            b"Signature: 8a477f597d28d172789f06886806bc55"
        )


_RUNTIME_CONTAINER_DIR_KEY: Optional[str] = None


def generated_content_dir(
    *,
    package: Optional["BinaryPackage"] = None,
    subdir_key: Optional[str] = None,
) -> str:
    global _RUNTIME_CONTAINER_DIR_KEY
    container_dir = _RUNTIME_CONTAINER_DIR_KEY
    first_run = False

    if container_dir is None:
        first_run = True
        container_dir = f"_pb-{os.getpid()}"
        _RUNTIME_CONTAINER_DIR_KEY = container_dir

    directory = os.path.join(scratch_dir(), container_dir)

    if first_run and os.path.isdir(directory):
        # In the unlikely case there is a re-run with exactly the same pid, `debputy` should not
        # see "stale" data.
        # TODO: Ideally, we would always clean up this directory on failure, but `atexit` is not
        #  reliable enough for that and we do not have an obvious hook for it.
        shutil.rmtree(directory)

    directory = os.path.join(
        directory,
        "generated-fs-content",
        f"pkg_{package.name}" if package else "no-package",
    )
    if subdir_key is not None:
        directory = os.path.join(directory, subdir_key)

    os.makedirs(directory, exist_ok=True)
    return directory


PerlConfigVars = collections.namedtuple(
    "PerlConfigVars", ["vendorlib", "vendorarch", "cross_inc_dir", "ld", "path_sep"]
)
PerlConfigData = collections.namedtuple("PerlConfigData", ["version", "debian_abi"])
_PERL_MODULE_DIRS: Dict[str, PerlConfigVars] = {}


@functools.lru_cache(1)
def _perl_config_data() -> PerlConfigData:
    d = (
        subprocess.check_output(
            [
                "perl",
                "-MConfig",
                "-e",
                'print "$Config{version}\n$Config{debian_abi}\n"',
            ]
        )
        .decode("utf-8")
        .splitlines()
    )
    return PerlConfigData(*d)


def _perl_version() -> str:
    return _perl_config_data().version


def perlxs_api_dependency() -> str:
    # dh_perl used the build version of perl for this, so we will too.  Most of the perl cross logic
    # assumes that the major version of build variant of Perl is the same as the host variant of Perl.
    config = _perl_config_data()
    if config.debian_abi is not None and config.debian_abi != "":
        return f"perlapi-{config.debian_abi}"
    return f"perlapi-{config.version}"


def resolve_perl_config(
    dpkg_architecture_variables: DpkgArchitectureBuildProcessValuesTable,
    dctrl_bin: Optional["BinaryPackage"],
) -> PerlConfigVars:
    global _PERL_MODULE_DIRS
    if dpkg_architecture_variables.is_cross_compiling:
        if dctrl_bin is not None:
            arch = dctrl_bin.resolved_architecture
            ma = dctrl_bin.deb_multiarch
        else:
            arch = dpkg_architecture_variables.current_host_arch
            ma = dpkg_architecture_variables.current_host_multiarch
    else:
        # We are de facto using the build-arch version of perl here; be explicit
        arch = "_build_arch_"
        ma = dpkg_architecture_variables["DEB_BUILD_MULTIARCH"]
    config_vars = _PERL_MODULE_DIRS.get(arch)
    if config_vars is None:
        cmd = ["perl"]
        if dpkg_architecture_variables.is_cross_compiling:
            version = _perl_version()
            cross_inc_dir = f"/usr/lib/{ma}/perl/cross-config-{version}"
            # FIXME: This should not fallback to "build-arch" but on the other hand, we use the perl module dirs
            #  for every package at the moment. So mandating correct perl dirs implies mandating perl-xs-dev in
            #  cross builds... meh.
            if os.path.exists(os.path.join(cross_inc_dir, "Config.pm")):
                cmd.append(f"-I{cross_inc_dir}")
        else:
            cross_inc_dir = None
        cmd.extend(
            [
                "-MConfig",
                "-e",
                'print "$Config{vendorlib}\n$Config{vendorarch}\n$Config{ld}\n$Config{path_sep}\n"',
            ]
        )
        output = subprocess.check_output(cmd).decode("utf-8").splitlines(keepends=False)
        if len(output) != 4:
            raise ValueError(
                "Internal error: Unable to determine the perl include directories:"
                f" Raw output from perl snippet: {output}"
            )
        config_vars = PerlConfigVars(
            vendorlib="/" + _normalize_path(output[0], with_prefix=False),
            vendorarch="/" + _normalize_path(output[1], with_prefix=False),
            cross_inc_dir=cross_inc_dir,
            ld=output[2],
            path_sep=output[3],
        )
        _PERL_MODULE_DIRS[arch] = config_vars
    return config_vars


@functools.lru_cache(1)
def detect_fakeroot() -> bool:
    if os.getuid() != 0 or "LD_PRELOAD" not in os.environ:
        return False
    env = dict(os.environ)
    del env["LD_PRELOAD"]
    try:
        return subprocess.check_output(["id", "-u"], env=env).strip() != b"0"
    except subprocess.CalledProcessError:
        print(
            'Could not run "id -u" with LD_PRELOAD unset; assuming we are not run under fakeroot',
            file=sys.stderr,
        )
        return False


@functools.lru_cache(1)
def _sc_arg_max() -> Optional[int]:
    try:
        return os.sysconf("SC_ARG_MAX")
    except RuntimeError:
        _warn("Could not resolve SC_ARG_MAX, falling back to a hard-coded limit")
        return None


def _split_xargs_args(
    static_cmd: Sequence[str],
    max_args_byte_len: int,
    varargs: Iterable[str],
    reuse_list_ok: bool,
) -> Iterator[List[str]]:
    static_cmd_len = len(static_cmd)
    remaining_len = max_args_byte_len
    pending_args = list(static_cmd)
    for arg in varargs:
        arg_len = len(arg.encode("utf-8")) + 1  # +1 for leading space
        remaining_len -= arg_len
        if not remaining_len:
            if len(pending_args) <= static_cmd_len:
                raise ValueError(
                    f"Could not fit a single argument into the command line !?"
                    f" {max_args_byte_len} (variable argument limit) < {arg_len} (argument length)"
                )
            yield pending_args
            remaining_len = max_args_byte_len - arg_len
            if reuse_list_ok:
                pending_args.clear()
                pending_args.extend(static_cmd)
            else:
                pending_args = list(static_cmd)
        pending_args.append(arg)

    if len(pending_args) > static_cmd_len:
        yield pending_args


def xargs(
    static_cmd: Sequence[str],
    varargs: Iterable[str],
    *,
    env: Optional[Mapping[str, str]] = None,
    reuse_list_ok: bool = False,
) -> Iterator[List[str]]:
    max_args_bytes = _sc_arg_max()
    # len overshoots with one space explaining the -1.  The _split_xargs_args
    # will account for the space for the first argument
    static_byte_len = (
        len(static_cmd) - 1 + sum(len(a.encode("utf-8")) for a in static_cmd)
    )
    if max_args_bytes is not None:
        if env is None:
            # +2 for nul bytes after key and value
            static_byte_len += sum(len(k) + len(v) + 2 for k, v in os.environb.items())
        else:
            # +2 for nul bytes after key and value
            static_byte_len += sum(
                len(k.encode("utf-8")) + len(v.encode("utf-8")) + 2
                for k, v in env.items()
            )
        # Add a fixed buffer for OS overhead here (in case env and cmd both must be page-aligned or something like
        # that)
        static_byte_len += 2 * 4096
    else:
        # The 20 000 limit is from debhelper, and it did not account for environment.  So neither will we here.
        max_args_bytes = 20_000
    remain_len = max_args_bytes - static_byte_len
    yield from _split_xargs_args(static_cmd, remain_len, varargs, reuse_list_ok)


# itertools recipe
def grouper(
    iterable: Iterable[T],
    n: int,
    *,
    incomplete: Literal["fill", "strict", "ignore"] = "fill",
    fillvalue: Optional[T] = None,
) -> Iterator[Tuple[T, ...]]:
    """Collect data into non-overlapping fixed-length chunks or blocks"""
    # grouper('ABCDEFG', 3, fillvalue='x') --> ABC DEF Gxx
    # grouper('ABCDEFG', 3, incomplete='strict') --> ABC DEF ValueError
    # grouper('ABCDEFG', 3, incomplete='ignore') --> ABC DEF
    args = [iter(iterable)] * n
    if incomplete == "fill":
        return zip_longest(*args, fillvalue=fillvalue)
    if incomplete == "strict":
        return zip(*args, strict=True)
    if incomplete == "ignore":
        return zip(*args)
    else:
        raise ValueError("Expected fill, strict, or ignore")


_LOGGING_SET_UP = False


def _check_color() -> Tuple[bool, bool, Optional[str]]:
    dpkg_or_default = os.environ.get(
        "DPKG_COLORS", "never" if "NO_COLOR" in os.environ else "auto"
    )
    requested_color = os.environ.get("DEBPUTY_COLORS", dpkg_or_default)
    bad_request = None
    if requested_color not in {"auto", "always", "never"}:
        bad_request = requested_color
        requested_color = "auto"

    if requested_color == "auto":
        stdout_color = sys.stdout.isatty()
        stderr_color = sys.stdout.isatty()
    else:
        enable = requested_color == "always"
        stdout_color = enable
        stderr_color = enable
    return stdout_color, stderr_color, bad_request


def program_name() -> str:
    name = os.path.basename(sys.argv[0])
    if name.endswith(".py"):
        name = name[:-3]
    if name == "__main__":
        name = os.path.basename(os.path.dirname(sys.argv[0]))
    # FIXME: Not optimal that we have to hardcode these kind of things here
    if name == "debputy_cmd":
        name = "debputy"
    return name


def package_cross_check_precheck(
    pkg_a: "BinaryPackage",
    pkg_b: "BinaryPackage",
) -> Tuple[bool, bool]:
    """Whether these two packages can do content cross-checks

    :param pkg_a: The first package
    :param pkg_b: The second package
    :return: A tuple if two booleans. If the first is True, then binary_package_a may do content cross-checks
      that invoĺves binary_package_b. If the second is True, then binary_package_b may do content cross-checks
      that involves binary_package_a. Both can be True and both can be False at the same time, which
      happens in common cases (arch:all + arch:any cases both to be False as a common example).
    """

    # Handle the two most obvious base-cases
    if not pkg_a.should_be_acted_on or not pkg_b.should_be_acted_on:
        return False, False
    if pkg_a.is_arch_all ^ pkg_b.is_arch_all:
        return False, False

    a_may_see_b = True
    b_may_see_a = True

    a_bp = pkg_a.fields.get("Build-Profiles", "")
    b_bp = pkg_b.fields.get("Build-Profiles", "")

    if a_bp != b_bp:
        a_bp_set = _parse_build_profiles(a_bp) if a_bp != "" else frozenset()
        b_bp_set = _parse_build_profiles(b_bp) if b_bp != "" else frozenset()

        # Check for build profiles being identically but just ordered differently.
        if a_bp_set != b_bp_set:
            # For simplicity, we let groups cancel each other out. If one side has no clauses
            # left, then it will always be built when the other is built.
            #
            # Eventually, someone will be here with a special case where more complex logic is
            # required. Good luck to you! Remember to add test cases for it (the existing logic
            # has some for a reason and if the logic is going to be more complex, it will need
            # tests cases to assert it fixes the problem and does not regress)
            if a_bp_set - b_bp_set:
                a_may_see_b = False
            if b_bp_set - a_bp_set:
                b_may_see_a = False

    if pkg_a.declared_architecture != pkg_b.declared_architecture:
        # Also here we could do a subset check, but wildcards vs. non-wildcards make that a pain
        if pkg_a.declared_architecture != "any":
            b_may_see_a = False
        if pkg_a.declared_architecture != "any":
            a_may_see_b = False

    return a_may_see_b, b_may_see_a


def change_log_level(
    log_level: int,
) -> None:
    if _DEFAULT_LOGGER is not None:
        _DEFAULT_LOGGER.setLevel(log_level)
    logging.getLogger("").setLevel(log_level)


def current_log_level() -> Optional[int]:
    if _DEFAULT_LOGGER is not None:
        return _DEFAULT_LOGGER.level
    return None


def setup_logging(
    *,
    log_only_to_stderr: bool = False,
    reconfigure_logging: bool = False,
) -> None:
    global _LOGGING_SET_UP, _DEFAULT_LOGGER, _STDOUT_HANDLER, _STDERR_HANDLER
    if _LOGGING_SET_UP and not reconfigure_logging:
        raise RuntimeError(
            "Logging has already been configured."
            " Use reconfigure_logging=True if you need to reconfigure it"
        )
    stdout_color, stderr_color, bad_request = _check_color()
    colors: Optional[Dict[str, str]] = None

    if stdout_color or stderr_color:
        try:
            import colorlog

        except ImportError:
            stdout_color = False
            stderr_color = False
        else:
            colors = dict(colorlog.default_log_colors)
            # Add our custom levels.
            colors["_INFO"] = colors["INFO"]
            colors["__INFO"] = colors["INFO"]

    if log_only_to_stderr:
        stdout = sys.stderr
        stdout_color = stderr_color
    else:
        stdout = sys.stderr

    class LogLevelFilter(logging.Filter):
        def __init__(self, threshold: int, above: bool):
            super().__init__()
            self.threshold = threshold
            self.above = above

        def filter(self, record: logging.LogRecord) -> bool:
            if self.above:
                return record.levelno >= self.threshold
            else:
                return record.levelno < self.threshold

    color_format = (
        "{bold}{name}{reset}: {bold}{log_color}{levelnamelower}{reset}: {message}"
    )
    colorless_format = "{name}: {levelnamelower}: {message}"

    existing_stdout_handler = _STDOUT_HANDLER
    existing_stderr_handler = _STDERR_HANDLER

    if stdout_color:
        stdout_handler = colorlog.StreamHandler(stdout)
        stdout_handler.setFormatter(
            colorlog.ColoredFormatter(
                color_format,
                style="{",
                force_color=True,
                log_colors=colors,
            )
        )
        logger = colorlog.getLogger()
        if existing_stdout_handler is not None:
            logger.removeHandler(existing_stdout_handler)
        _STDOUT_HANDLER = stdout_handler
        logger.addHandler(stdout_handler)
    else:
        stdout_handler = logging.StreamHandler(stdout)
        stdout_handler.setFormatter(logging.Formatter(colorless_format, style="{"))
        logger = logging.getLogger()
        if existing_stdout_handler is not None:
            logger.removeHandler(existing_stdout_handler)
        _STDOUT_HANDLER = stdout_handler
        logger.addHandler(stdout_handler)

    if stderr_color:
        stderr_handler = colorlog.StreamHandler(sys.stderr)
        stderr_handler.setFormatter(
            colorlog.ColoredFormatter(
                color_format,
                style="{",
                force_color=True,
                log_colors=colors,
            )
        )
        logger = logging.getLogger()
        if existing_stderr_handler is not None:
            logger.removeHandler(existing_stderr_handler)
        _STDERR_HANDLER = stderr_handler
        logger.addHandler(stderr_handler)
    else:
        stderr_handler = logging.StreamHandler(sys.stderr)
        stderr_handler.setFormatter(logging.Formatter(colorless_format, style="{"))
        logger = logging.getLogger()
        if existing_stderr_handler is not None:
            logger.removeHandler(existing_stderr_handler)
        _STDERR_HANDLER = stderr_handler
        logger.addHandler(stderr_handler)

    stdout_handler.addFilter(LogLevelFilter(logging.WARN, False))
    stderr_handler.addFilter(LogLevelFilter(logging.WARN, True))

    name = program_name()

    old_factory = logging.getLogRecordFactory()

    def record_factory(
        *args: Any, **kwargs: Any
    ) -> logging.LogRecord:  # pragma: no cover
        record = old_factory(*args, **kwargs)
        record.levelname = record.levelname.lstrip("_")
        record.levelnamelower = record.levelname.lower()
        return record

    logging.setLogRecordFactory(record_factory)

    logging.getLogger().setLevel(logging.WARN)
    _DEFAULT_LOGGER = logging.getLogger(name)

    if bad_request:
        _DEFAULT_LOGGER.warning(
            f'Invalid color request for "{bad_request}" in either DEBPUTY_COLORS or DPKG_COLORS.'
            ' Resetting to "auto".'
        )

    _LOGGING_SET_UP = True



import itertools
from typing import Any, Optional, Tuple, IO, Iterable

from debputy.plugin.api import DebputyPluginInitializer, VirtualPath

_BUF_SIZE = 8096


def initialize(api: DebputyPluginInitializer) -> None:
    register_processing_steps_via_non_public_api(api)


def register_processing_steps_via_non_public_api(api: Any) -> None:
    api.package_processor(
        "deduplicate-lua-files",
        deduplicate_lua_files,
    )


def find_lowest_lua5_version_dir(fs_root: VirtualPath) -> Optional[Tuple[int, VirtualPath]]:
    for lua5_minor_version in range(1, 9):
        lua_dir = fs_root.lookup(f"/usr/share/lua/5.{lua5_minor_version}/")
        if not lua_dir:
            continue
        return lua5_minor_version, lua_dir
    return None


def _compute_rel_path(reference_point: VirtualPath, target_file: VirtualPath) -> str:
    assert target_file.path.startswith(reference_point.path + "/")
    return target_file.path[len(reference_point.path) + 1:]


def _chunk(fd: IO[bytes]) -> Iterable[bytes]:
    c = fd.read(_BUF_SIZE)
    while c:
        yield c
        c = fd.read(_BUF_SIZE)


def _same_content(a: VirtualPath, b: VirtualPath) -> bool:
    if a.size != b.size:
        return False
    # It might be tempting to use `cmp {a.fs_path} {b.fs_path}`, but this method makes it easier to test
    # as the test will not have to create physical files.
    #
    # We could also replace this with a checksum check, then we could store them in a table and deduplicate
    # between multiple versions (not just against the lowest version). Decided against it because this is
    # are more faithful translation of the original code.
    with a.open(byte_io=True, buffering=_BUF_SIZE) as afd, b.open(byte_io=True, buffering=_BUF_SIZE) as bfd:
        for a_chunk, b_chunk in itertools.zip_longest(afd, bfd, fillvalue=None):
            if a_chunk is None or b_chunk is None or a_chunk != b_chunk:
                return False
    return True


def deduplicate_lua_files(fs_root: VirtualPath, _unused1: Any, _unused2: Any) -> None:
    res_tuple = find_lowest_lua5_version_dir(fs_root)
    if res_tuple is None:
        return
    lua5_minor_version, lua5_reference_dir = res_tuple
    all_lua5_ref_files = [
        lua_file
        for lua_file in lua5_reference_dir.all_paths()
        if lua_file.is_file and lua_file.name.endswith(".lua")
    ]
    for lua_file in all_lua5_ref_files:
        rel_path = _compute_rel_path(lua5_reference_dir, lua_file)
        for newer_lua_minor_version in range(lua5_minor_version + 1, 9):
            alt_lua_file = fs_root.lookup(f"/usr/share/lua/5.{newer_lua_minor_version}/{rel_path}")
            if alt_lua_file is None or not alt_lua_file.is_file or not _same_content(lua_file, alt_lua_file):
                continue
            parent_dir = alt_lua_file.parent_dir
            assert parent_dir is not None
            alt_lua_file.unlink()
            # We do not have to normalize the link; debputy will handle that later.
            parent_dir.add_symlink(alt_lua_file.name, lua_file.absolute)



import dataclasses
import os
import subprocess
from functools import lru_cache
from typing import (
    Sequence,
    Optional,
    Union,
    Literal,
    Tuple,
    Mapping,
    Iterable,
    TYPE_CHECKING,
    Dict,
    MutableMapping,
    NotRequired,
)

from debputy.manifest_conditions import ManifestCondition
from debputy.manifest_parser.exceptions import ManifestParseException
from debputy.manifest_parser.tagging_types import DebputyParsedContent
from debputy.manifest_parser.util import (
    AttributePath,
    _SymbolicModeSegment,
    parse_symbolic_mode,
)
from debputy.path_matcher import MatchRule, ExactFileSystemPath
from debputy.substitution import Substitution
from debputy.util import _normalize_path, _error, _warn, _debug_log

if TYPE_CHECKING:
    from debputy.manifest_parser.parser_data import ParserContextData


@dataclasses.dataclass(slots=True, frozen=True)
class OwnershipDefinition:
    entity_name: str
    entity_id: int


class DebputyParsedContentStandardConditional(DebputyParsedContent):
    when: NotRequired[ManifestCondition]


ROOT_DEFINITION = OwnershipDefinition("root", 0)


BAD_OWNER_NAMES = {
    "_apt",  # All things owned by _apt are generated by apt after installation
    "nogroup",  # It is not supposed to own anything as it is an entity used for dropping permissions
    "nobody",  # It is not supposed to own anything as it is an entity used for dropping permissions
}
BAD_OWNER_IDS = {
    65534,  # ID of nobody / nogroup
}


def _parse_ownership(
    v: Union[str, int],
    attribute_path: AttributePath,
) -> Tuple[Optional[str], Optional[int]]:
    if isinstance(v, str) and ":" in v:
        if v == ":":
            raise ManifestParseException(
                f'Invalid ownership value "{v}" at {attribute_path.path}: Ownership is redundant if it is ":"'
                f" (blank name and blank id).  Please provide non-default values or remove the definition."
            )
        entity_name: Optional[str]
        entity_id: Optional[int]
        entity_name, entity_id_str = v.split(":")
        if entity_name == "":
            entity_name = None
        if entity_id_str != "":
            entity_id = int(entity_id_str)
        else:
            entity_id = None
        return entity_name, entity_id

    if isinstance(v, int):
        return None, v
    if v.isdigit():
        raise ManifestParseException(
            f'Invalid ownership value "{v}" at {attribute_path.path}: The provided value "{v}" is a string (implying'
            " name lookup), but it contains an integer (implying id lookup).  Please use a regular int for id lookup"
            f' (removing the quotes) or add a ":" in the end ("{v}:") as a disambiguation if you are *really* looking'
            " for an entity with that name."
        )
    return v, None


@lru_cache
def _load_ownership_table_from_file(
    name: Literal["passwd.master", "group.master"],
) -> Tuple[Mapping[str, OwnershipDefinition], Mapping[int, OwnershipDefinition]]:
    filename = os.path.join("/usr/share/base-passwd", name)
    name_table = {}
    uid_table = {}
    for owner_def in _read_ownership_def_from_base_password_template(filename):
        # Could happen if base-passwd template has two users with the same ID.  We assume this will not occur.
        assert owner_def.entity_name not in name_table
        assert owner_def.entity_id not in uid_table
        name_table[owner_def.entity_name] = owner_def
        uid_table[owner_def.entity_id] = owner_def

    return name_table, uid_table


def _read_ownership_def_from_base_password_template(
    template_file: str,
) -> Iterable[OwnershipDefinition]:
    with open(template_file) as fd:
        for line in fd:
            entity_name, _star, entity_id, _remainder = line.split(":", 3)
            if entity_id == "0" and entity_name == "root":
                yield ROOT_DEFINITION
            else:
                yield OwnershipDefinition(entity_name, int(entity_id))


class FileSystemMode:
    @classmethod
    def parse_filesystem_mode(
        cls,
        mode_raw: str,
        attribute_path: AttributePath,
    ) -> "FileSystemMode":
        if mode_raw and mode_raw[0].isdigit():
            return OctalMode.parse_filesystem_mode(mode_raw, attribute_path)
        return SymbolicMode.parse_filesystem_mode(mode_raw, attribute_path)

    def compute_mode(self, current_mode: int, is_dir: bool) -> int:
        raise NotImplementedError


@dataclasses.dataclass(slots=True, frozen=True)
class SymbolicMode(FileSystemMode):
    provided_mode: str
    segments: Sequence[_SymbolicModeSegment]

    @classmethod
    def parse_filesystem_mode(
        cls,
        mode_raw: str,
        attribute_path: AttributePath,
    ) -> "SymbolicMode":
        segments = list(parse_symbolic_mode(mode_raw, attribute_path))
        return SymbolicMode(mode_raw, segments)

    def __str__(self) -> str:
        return self.symbolic_mode()

    @property
    def is_symbolic_mode(self) -> bool:
        return False

    def symbolic_mode(self) -> str:
        return self.provided_mode

    def compute_mode(self, current_mode: int, is_dir: bool) -> int:
        final_mode = current_mode
        for segment in self.segments:
            final_mode = segment.apply(final_mode, is_dir)
        return final_mode


@dataclasses.dataclass(slots=True, frozen=True)
class OctalMode(FileSystemMode):
    octal_mode: int

    @classmethod
    def parse_filesystem_mode(
        cls,
        mode_raw: str,
        attribute_path: AttributePath,
    ) -> "FileSystemMode":
        try:
            mode = int(mode_raw, base=8)
        except ValueError as e:
            error_msg = 'An octal mode must be all digits between 0-7 (such as "644")'
            raise ManifestParseException(
                f"Cannot parse {attribute_path.path} as an octal mode: {error_msg}"
            ) from e
        return OctalMode(mode)

    @property
    def is_octal_mode(self) -> bool:
        return True

    def compute_mode(self, _current_mode: int, _is_dir: bool) -> int:
        return self.octal_mode

    def __str__(self) -> str:
        return f"0{oct(self.octal_mode)[2:]}"


@dataclasses.dataclass(slots=True, frozen=True)
class _StaticFileSystemOwnerGroup:
    ownership_definition: OwnershipDefinition

    @property
    def entity_name(self) -> str:
        return self.ownership_definition.entity_name

    @property
    def entity_id(self) -> int:
        return self.ownership_definition.entity_id

    @classmethod
    def from_manifest_value(
        cls,
        raw_input: Union[str, int],
        attribute_path: AttributePath,
    ) -> "_StaticFileSystemOwnerGroup":
        provided_name, provided_id = _parse_ownership(raw_input, attribute_path)
        owner_def = cls._resolve(raw_input, provided_name, provided_id, attribute_path)
        if (
            owner_def.entity_name in BAD_OWNER_NAMES
            or owner_def.entity_id in BAD_OWNER_IDS
        ):
            raise ManifestParseException(
                f'Refusing to use "{raw_input}" as {cls._owner_type()} (defined at {attribute_path.path})'
                f' as it resolves to "{owner_def.entity_name}:{owner_def.entity_id}" and no path should have this'
                f" entity as {cls._owner_type()} as it is unsafe."
            )
        return cls(owner_def)

    @classmethod
    def _resolve(
        cls,
        raw_input: Union[str, int],
        provided_name: Optional[str],
        provided_id: Optional[int],
        attribute_path: AttributePath,
    ) -> OwnershipDefinition:
        table_name = cls._ownership_table_name()
        name_table, id_table = _load_ownership_table_from_file(table_name)
        name_match = (
            name_table.get(provided_name) if provided_name is not None else None
        )
        id_match = id_table.get(provided_id) if provided_id is not None else None
        if id_match is None and name_match is None:
            name_part = provided_name if provided_name is not None else "N/A"
            id_part = provided_id if provided_id is not None else "N/A"
            raise ManifestParseException(
                f'Cannot resolve "{raw_input}" as {cls._owner_type()} (from {attribute_path.path}):'
                f" It is not known to be a static {cls._owner_type()} from base-passwd."
                f' The value was interpreted as name: "{name_part}" and id: {id_part}'
            )
        if id_match is None:
            assert name_match is not None
            return name_match
        if name_match is None:
            assert id_match is not None
            return id_match
        if provided_name != id_match.entity_name:
            raise ManifestParseException(
                f"Bad {cls._owner_type()} declaration: The id {provided_id} resolves to {id_match.entity_name}"
                f" according to base-passwd, but the packager declared to should have been {provided_name}"
                f" at {attribute_path.path}"
            )
        if provided_id != name_match.entity_id:
            raise ManifestParseException(
                f"Bad {cls._owner_type} declaration: The name {provided_name} resolves to {name_match.entity_id}"
                f" according to base-passwd, but the packager declared to should have been {provided_id}"
                f" at {attribute_path.path}"
            )
        return id_match

    @classmethod
    def _owner_type(cls) -> Literal["owner", "group"]:
        raise NotImplementedError

    @classmethod
    def _ownership_table_name(cls) -> Literal["passwd.master", "group.master"]:
        raise NotImplementedError


class StaticFileSystemOwner(_StaticFileSystemOwnerGroup):
    @classmethod
    def _owner_type(cls) -> Literal["owner", "group"]:
        return "owner"

    @classmethod
    def _ownership_table_name(cls) -> Literal["passwd.master", "group.master"]:
        return "passwd.master"


class StaticFileSystemGroup(_StaticFileSystemOwnerGroup):
    @classmethod
    def _owner_type(cls) -> Literal["owner", "group"]:
        return "group"

    @classmethod
    def _ownership_table_name(cls) -> Literal["passwd.master", "group.master"]:
        return "group.master"


@dataclasses.dataclass(slots=True, frozen=True)
class SymlinkTarget:
    raw_symlink_target: str
    attribute_path: AttributePath
    symlink_target: str

    @classmethod
    def parse_symlink_target(
        cls,
        raw_symlink_target: str,
        attribute_path: AttributePath,
        substitution: Substitution,
    ) -> "SymlinkTarget":
        return SymlinkTarget(
            raw_symlink_target,
            attribute_path,
            substitution.substitute(raw_symlink_target, attribute_path.path),
        )


class FileSystemMatchRule:
    @property
    def raw_match_rule(self) -> str:
        raise NotImplementedError

    @property
    def attribute_path(self) -> AttributePath:
        raise NotImplementedError

    @property
    def match_rule(self) -> MatchRule:
        raise NotImplementedError

    @classmethod
    def parse_path_match(
        cls,
        raw_match_rule: str,
        attribute_path: AttributePath,
        parser_context: "ParserContextData",
    ) -> "FileSystemMatchRule":
        return cls.from_path_match(
            raw_match_rule, attribute_path, parser_context.substitution
        )

    @classmethod
    def from_path_match(
        cls,
        raw_match_rule: str,
        attribute_path: AttributePath,
        substitution: "Substitution",
    ) -> "FileSystemMatchRule":
        try:
            mr = MatchRule.from_path_or_glob(
                raw_match_rule,
                attribute_path.path,
                substitution=substitution,
            )
        except ValueError as e:
            raise ManifestParseException(
                f'Could not parse "{raw_match_rule}" (defined at {attribute_path.path})'
                f" as a path or a glob: {e.args[0]}"
            )

        if isinstance(mr, ExactFileSystemPath):
            return FileSystemExactMatchRule(
                raw_match_rule,
                attribute_path,
                mr,
            )
        return FileSystemGenericMatch(
            raw_match_rule,
            attribute_path,
            mr,
        )


@dataclasses.dataclass(slots=True, frozen=True)
class FileSystemGenericMatch(FileSystemMatchRule):
    raw_match_rule: str
    attribute_path: AttributePath
    match_rule: MatchRule


@dataclasses.dataclass(slots=True, frozen=True)
class FileSystemExactMatchRule(FileSystemMatchRule):
    raw_match_rule: str
    attribute_path: AttributePath
    match_rule: ExactFileSystemPath

    @classmethod
    def from_path_match(
        cls,
        raw_match_rule: str,
        attribute_path: AttributePath,
        substitution: "Substitution",
    ) -> "FileSystemExactMatchRule":
        try:
            normalized = _normalize_path(raw_match_rule)
        except ValueError as e:
            raise ManifestParseException(
                f'The path "{raw_match_rule}" provided in {attribute_path.path} should be relative to the'
                ' root of the package and not use any ".." or "." segments.'
            ) from e
        if normalized == ".":
            raise ManifestParseException(
                f'The path "{raw_match_rule}" matches a file system root and that is not a valid match'
                f' at "{attribute_path.path}". Please narrow the provided path.'
            )
        mr = ExactFileSystemPath(
            substitution.substitute(normalized, attribute_path.path)
        )
        if mr.path.endswith("/") and issubclass(cls, FileSystemExactNonDirMatchRule):
            raise ManifestParseException(
                f'The path "{raw_match_rule}" at {attribute_path.path} resolved to'
                f' "{mr.path}". Since the resolved path ends with a slash ("/"), this'
                " means only a directory can match. However, this attribute should"
                " match a *non*-directory"
            )
        return cls(
            raw_match_rule,
            attribute_path,
            mr,
        )


class FileSystemExactNonDirMatchRule(FileSystemExactMatchRule):
    pass


class BuildEnvironmentDefinition:

    def dpkg_buildflags_env(
        self,
        env: Mapping[str, str],
        definition_source: Optional[str],
    ) -> Dict[str, str]:
        dpkg_env = {}
        try:
            bf_output = subprocess.check_output(["dpkg-buildflags"], env=env)
        except FileNotFoundError:
            if definition_source is None:
                _error(
                    "The dpkg-buildflags command was not available and is necessary to set the relevant"
                    "env variables by default."
                )
            _error(
                "The dpkg-buildflags command was not available and is necessary to set the relevant"
                f"env variables for the environment defined at {definition_source}."
            )
        except subprocess.CalledProcessError as e:
            if definition_source is None:
                _error(
                    f"The dpkg-buildflags command failed with exit code {e.returncode}. Please review the output from"
                    f" dpkg-buildflags above to resolve the issue."
                )
            _error(
                f"The dpkg-buildflags command failed with exit code {e.returncode}. Please review the output from"
                f" dpkg-buildflags above to resolve the issue. The environment definition that triggered this call"
                f" was {definition_source}"
            )
        else:
            warned = False
            for line in bf_output.decode("utf-8").splitlines(keepends=False):
                if "=" not in line or line.startswith("="):
                    if not warned:
                        _warn(
                            f"Unexpected output from dpkg-buildflags (not a K=V line): {line}"
                        )
                    continue
                k, v = line.split("=", 1)
                if k.strip() != k:
                    if not warned:
                        _warn(
                            f'Unexpected output from dpkg-buildflags (Key had spaces): "{line}"'
                        )
                    continue
                dpkg_env[k] = v

        return dpkg_env

    def log_computed_env(self, source: str, computed_env: Mapping[str, str]) -> None:
        _debug_log(f"Computed environment variables from {source}")
        for k, v in computed_env.items():
            _debug_log(f"  {k}={v}")

    def update_env(self, env: MutableMapping[str, str]) -> None:
        dpkg_env = self.dpkg_buildflags_env(env, None)
        self.log_computed_env("dpkg-buildflags", dpkg_env)
        env.update(dpkg_env)


class BuildEnvironments:

    def __init__(
        self,
        environments: Dict[str, BuildEnvironmentDefinition],
        default_environment: Optional[BuildEnvironmentDefinition],
    ) -> None:
        self.environments = environments
        self.default_environment = default_environment



from debputy.exceptions import DebputyRuntimeError


class ManifestException(DebputyRuntimeError):
    pass


class ManifestParseException(ManifestException):
    pass


class ManifestTypeException(ManifestParseException):
    pass


class ManifestInvalidUserDataException(ManifestException):
    pass



from typing import Mapping, Optional, TYPE_CHECKING

from debputy.architecture_support import DpkgArchitectureBuildProcessValuesTable
from debputy.manifest_conditions import _run_build_time_tests

if TYPE_CHECKING:
    from debputy.commands.debputy_cmd.context import CommandContext


class BuildContext:
    @staticmethod
    def from_command_context(
        cmd_context: "CommandContext",
    ) -> "BuildContext":
        return BuildContextImpl(cmd_context)

    @property
    def deb_build_options(self) -> Mapping[str, Optional[str]]:
        raise NotImplementedError

    def parallelization_limit(self, *, support_zero_as_unlimited: bool = False) -> int:
        """Parallelization limit of the build

        This is accessor that reads the `parallel` option from `DEB_BUILD_OPTIONS` with relevant
        fallback behavior.

        :param support_zero_as_unlimited: The debhelper framework allowed `0` to mean unlimited
          in some build systems. If the build system supports this, it should set this option
          to True, which will allow `0` as a possible return value. WHen this option is False
          (which is the default), `0` will be remapped to a high number to preserve the effect
          in spirit (said fallback number is also from `debhelper`).
        """
        limit = self.deb_build_options.get("parallel")
        if limit is None:
            return 1
        try:
            v = int(limit)
        except ValueError:
            return 1
        if v == 0 and not support_zero_as_unlimited:
            # debhelper allowed "0" to be used as unlimited in some cases. Preserve that feature
            # for callers that are prepared for it. For everyone else, remap 0 to an obscene number
            # that de facto has the same behavior
            #
            # The number is taken out of `cmake.pm` from `debhelper` to be "Bug compatible" with
            # debhelper on the fallback as well.
            return 999
        return v

    @property
    def is_terse_build(self) -> bool:
        """Whether the build is terse

        This is a shorthand for testing for `terse` in DEB_BUILD_OPTIONS
        """
        return "terse" in self.deb_build_options

    @property
    def is_cross_compiling(self) -> bool:
        """Whether the build is considered a cross build

        Note: Do **not** use this as indicator for whether tests should run. Use `should_run_tests` instead.
          To the naive eye, they seem like they overlap in functionality, but they do not. There are cross
          builds where tests can be run. Additionally, there are non-cross-builds where tests should be
          skipped.
        """
        return self.dpkg_architecture_variables.is_cross_compiling

    def cross_tool(self, command: str) -> str:
        if not self.is_cross_compiling:
            return command
        cross_prefix = self.dpkg_architecture_variables["DEB_HOST_GNU_TYPE"]
        return f"{cross_prefix}-{command}"

    @property
    def dpkg_architecture_variables(self) -> DpkgArchitectureBuildProcessValuesTable:
        raise NotImplementedError

    @property
    def should_run_tests(self) -> bool:
        return _run_build_time_tests(self.deb_build_options)


class BuildContextImpl(BuildContext):
    def __init__(
        self,
        cmd_context: "CommandContext",
    ) -> None:
        self._cmd_context = cmd_context

    @property
    def deb_build_options(self) -> Mapping[str, Optional[str]]:
        return self._cmd_context.deb_build_options

    @property
    def dpkg_architecture_variables(self) -> DpkgArchitectureBuildProcessValuesTable:
        return self._cmd_context.dpkg_architecture_variables()



import dataclasses
import os.path
import subprocess
from typing import Callable, Sequence, List, Optional, TypeVar

from debian.debian_support import Version

from debputy.util import _error


@dataclasses.dataclass(slots=True, frozen=True)
class LSPSelfCheck:
    feature: str
    test: Callable[[], bool]
    problem: str
    how_to_fix: str
    is_mandatory: bool = False


LSP_CHECKS: List[LSPSelfCheck] = []

C = TypeVar("C", bound="Callable")


def lsp_import_check(
    packages: Sequence[str],
    *,
    feature_name: Optional[str] = None,
    is_mandatory: bool = False,
) -> Callable[[C], C]:

    def _wrapper(func: C) -> C:

        def _impl():
            try:
                r = func()
            except ImportError:
                return False
            return r is None or bool(r)

        suffix = "fix this issue" if is_mandatory else "enable this feature"

        LSP_CHECKS.append(
            LSPSelfCheck(
                _feature_name(feature_name, func),
                _impl,
                "Missing dependencies",
                f"Run `apt satisfy '{', '.join(packages)}'` to {suffix}",
                is_mandatory=is_mandatory,
            )
        )
        return func

    return _wrapper


def lsp_generic_check(
    problem: str,
    how_to_fix: str,
    *,
    feature_name: Optional[str] = None,
    is_mandatory: bool = False,
) -> Callable[[C], C]:

    def _wrapper(func: C) -> C:
        LSP_CHECKS.append(
            LSPSelfCheck(
                _feature_name(feature_name, func),
                func,
                problem,
                how_to_fix,
                is_mandatory=is_mandatory,
            )
        )
        return func

    return _wrapper


def _feature_name(feature: Optional[str], func: Callable[[], None]) -> str:
    if feature is not None:
        return feature
    return func.__name__.replace("_", " ")


@lsp_import_check(["python3-lsprotocol", "python3-pygls"], is_mandatory=True)
def minimum_requirements() -> bool:
    import pygls.server

    # The hasattr is irrelevant; but it avoids the import being flagged as redundant.
    return hasattr(pygls.server, "LanguageServer")


@lsp_import_check(["python3-levenshtein"])
def typo_detection() -> bool:
    import Levenshtein

    # The hasattr is irrelevant; but it avoids the import being flagged as redundant.
    return hasattr(Levenshtein, "distance")


@lsp_import_check(["hunspell-en-us", "python3-hunspell"])
def spell_checking() -> bool:
    import Levenshtein

    # The hasattr is irrelevant; but it avoids the import being flagged as redundant.
    return hasattr(Levenshtein, "distance") and os.path.exists(
        "/usr/share/hunspell/en_US.dic"
    )


@lsp_generic_check(
    feature_name="extra dh support",
    problem="Missing dependencies",
    how_to_fix="Run `apt satisfy debhelper (>= 13.16~)` to enable this feature",
)
def check_dh_version() -> bool:
    try:
        output = subprocess.check_output(
            [
                "dpkg-query",
                "-W",
                "--showformat=${Version} ${db:Status-Status}\n",
                "debhelper",
            ]
        ).decode("utf-8")
    except (FileNotFoundError, subprocess.CalledProcessError):
        return False
    else:
        parts = output.split()
        if len(parts) != 2:
            return False
        if parts[1] != "installed":
            return False
        return Version(parts[0]) >= Version("13.16~")


@lsp_generic_check(
    feature_name="apt cache packages",
    problem="Missing apt or empty apt cache",
    how_to_fix="",
)
def check_apt_cache() -> bool:
    try:
        output = subprocess.check_output(
            [
                "apt-get",
                "indextargets",
                "--format",
                "$(IDENTIFIER)",
            ]
        ).decode("utf-8")
    except (FileNotFoundError, subprocess.CalledProcessError):
        return False
    for line in output.splitlines():
        if line.strip() == "Packages":
            return True

    return False


def assert_can_start_lsp() -> None:
    for self_check in LSP_CHECKS:
        if self_check.is_mandatory and not self_check.test():
            _error(
                f"Cannot start the language server. {self_check.problem}. {self_check.how_to_fix}"
            )



from typing import cast, TYPE_CHECKING

if TYPE_CHECKING:
    from debputy.plugin.api.impl_types import DebputyPluginMetadata


class DebputyRuntimeError(RuntimeError):
    @property
    def message(self) -> str:
        return cast("str", self.args[0])


class DebputyRuntimeErrorWithPreamble(DebputyRuntimeError):

    def render_preamble(self) -> None:
        raise NotImplementedError


class DebputyBuildStepError(DebputyRuntimeError):
    pass


class DebputySubstitutionError(DebputyRuntimeError):
    pass


class DebputyManifestVariableRequiresDebianDirError(DebputySubstitutionError):
    pass


class DebputyDpkgGensymbolsError(DebputyRuntimeError):
    pass


class SymlinkLoopError(ValueError):
    @property
    def message(self) -> str:
        return cast("str", self.args[0])


class PureVirtualPathError(TypeError):
    @property
    def message(self) -> str:
        return cast("str", self.args[0])


class TestPathWithNonExistentFSPathError(TypeError):
    @property
    def message(self) -> str:
        return cast("str", self.args[0])


class DebputyFSError(DebputyRuntimeError):
    pass


class DebputyFSIsROError(DebputyFSError):
    pass


class PluginBaseError(DebputyRuntimeError):
    pass


class DebputyPluginRuntimeError(PluginBaseError):
    pass


class PluginNotFoundError(PluginBaseError):
    pass


class PluginInitializationError(PluginBaseError):
    pass


class PluginIncorrectRegistrationError(PluginInitializationError):
    pass


class PluginMetadataError(PluginBaseError):
    pass


class PluginConflictError(PluginBaseError):
    @property
    def plugin_a(self) -> "DebputyPluginMetadata":
        return cast("DebputyPluginMetadata", self.args[1])

    @property
    def plugin_b(self) -> "DebputyPluginMetadata":
        return cast("DebputyPluginMetadata", self.args[2])


class PluginAPIViolationError(PluginBaseError):
    pass


class UnhandledOrUnexpectedErrorFromPluginError(PluginBaseError):
    pass


class DebputyMetadataAccessError(DebputyPluginRuntimeError):
    pass



from typing import Optional

from debputy.plugin.api import virtual_path_def, VirtualPath
from debputy.plugin.api.test_api import (
    initialize_plugin_under_test,
    build_virtual_file_system,
)


def test_lua_plugin_deduplicate_lua_files_no_lua() -> None:
    plugin = initialize_plugin_under_test()
    no_lua_files_fs = build_virtual_file_system(
        [
            "usr/bin/foo",
            "usr/share/lua5.2/foo/",  # Said no lua files, not no lua paths :)
        ]
    )
    path_count = sum((1 for _ in no_lua_files_fs.all_paths()), start=0)
    plugin.run_package_processor("deduplicate-lua-files", no_lua_files_fs)
    # The plugin does nothing if there are no lua files
    assert path_count == sum((1 for _ in no_lua_files_fs.all_paths()), start=0)
    assert not any(f.is_symlink for f in no_lua_files_fs.all_paths())


def _resolve_lua_symlink(v: VirtualPath) -> Optional[VirtualPath]:
    # TODO: This should probably be in the VirtualPath API (after some corner case tests/improvements)
    assert v.is_symlink
    target = v.readlink()
    parent = v.parent_dir
    assert parent is not None
    return parent.lookup(target)


def test_lua_plugin_deduplicate_lua_files_lua_differ() -> None:
    plugin = initialize_plugin_under_test()
    # Long to test more than one iteration
    base_content = "a" * 10000
    fs = build_virtual_file_system(
        [
            virtual_path_def("usr/share/lua/5.2/foo.lua", content=base_content),
            virtual_path_def("usr/share/lua/5.3/foo.lua", content=base_content),
            virtual_path_def("usr/share/lua/5.4/foo.lua", content=base_content),
            virtual_path_def("usr/share/lua/5.2/bar.lua", content=base_content + "unique for 5.2"),
            virtual_path_def("usr/share/lua/5.3/bar.lua", content=base_content + "unique for 5.3"),
            # Deliberately shorter to test different parse lengths
            virtual_path_def("usr/share/lua/5.4/bar.lua", content="unique for 5.4"),
            virtual_path_def("usr/share/lua/5.2/foo5.2.lua", content="unrelated"),
            virtual_path_def("usr/share/lua/5.3/foo5.3.lua", content="unrelated"),
            virtual_path_def("usr/share/lua/5.4/foo5.4.lua", content="unrelated"),
        ]
    )
    plugin.run_package_processor("deduplicate-lua-files", fs)

    foo52_lua = fs.lookup("/usr/share/lua/5.2/foo.lua")
    # This should be there and unchanged
    assert foo52_lua is not None and foo52_lua.is_file
    # These should be symlinks to foo52_lua
    for foo5x_lua_path in ["/usr/share/lua/5.3/foo.lua", "/usr/share/lua/5.4/foo.lua"]:
        foo5x_lua = fs.lookup(foo5x_lua_path)
        assert foo5x_lua_path is not None and foo5x_lua.is_symlink
        target_path = _resolve_lua_symlink(foo5x_lua)
        assert target_path.path == foo52_lua.path

    # These should be unchanged
    for path_stem in [
        "5.2/bar.lua",
        "5.3/bar.lua",
        "5.4/bar.lua",
        "5.2/foo5.2.lua",
        "5.3/foo5.3.lua",
        "5.4/foo5.4.lua",
    ]:
        unique_file = fs.lookup(f"/usr/share/lua/{path_stem}")
        assert unique_file is not None and unique_file.is_file



import pathlib

from .version import IS_RELEASE_BUILD, __version__, _RUN_FROM_SOURCE

# Replaced during install; must be a single line
# fmt: off
DEBPUTY_ROOT_DIR = pathlib.Path(__file__).parent.parent.parent
DEBPUTY_PLUGIN_ROOT_DIR = pathlib.Path(__file__).parent.parent.parent
# fmt: on

if IS_RELEASE_BUILD:
    DEBPUTY_DOC_ROOT_DIR = (
        f"https://salsa.debian.org/debian/debputy/-/blob/debian/{__version__}"
    )
else:
    DEBPUTY_DOC_ROOT_DIR = "https://salsa.debian.org/debian/debputy/-/blob/main"

DEBPUTY_IS_RUN_FROM_SOURCE = _RUN_FROM_SOURCE



import contextlib
import dataclasses
import inspect
import os.path
from importlib.resources.abc import Traversable
from io import BytesIO
from pathlib import Path
from typing import (
    Mapping,
    Dict,
    Optional,
    Tuple,
    List,
    cast,
    FrozenSet,
    Sequence,
    Union,
    Type,
    Iterator,
    Set,
    KeysView,
    Callable,
)

from debian.deb822 import Deb822
from debian.substvars import Substvars

from debputy import DEBPUTY_PLUGIN_ROOT_DIR
from debputy.architecture_support import faked_arch_table
from debputy.filesystem_scan import FSROOverlay, FSRootDir
from debputy.packages import BinaryPackage
from debputy.plugin.api import (
    PluginInitializationEntryPoint,
    VirtualPath,
    PackageProcessingContext,
    DpkgTriggerType,
    Maintscript,
)
from debputy.plugin.api.example_processing import process_discard_rule_example
from debputy.plugin.api.impl import (
    plugin_metadata_for_debputys_own_plugin,
    DebputyPluginInitializerProvider,
    parse_json_plugin_desc,
    MaintscriptAccessorProviderBase,
    BinaryCtrlAccessorProviderBase,
    PLUGIN_TEST_SUFFIX,
    find_json_plugin,
    ServiceDefinitionImpl,
)
from debputy.plugin.api.impl_types import (
    PackagerProvidedFileClassSpec,
    DebputyPluginMetadata,
    PluginProvidedTrigger,
    ServiceManagerDetails,
)
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from debputy.plugin.api.spec import (
    MaintscriptAccessor,
    FlushableSubstvars,
    ServiceRegistry,
    DSD,
    ServiceUpgradeRule,
)
from debputy.plugin.api.test_api.test_spec import (
    InitializedPluginUnderTest,
    RegisteredPackagerProvidedFile,
    RegisteredTrigger,
    RegisteredMaintscript,
    DEBPUTY_TEST_AGAINST_INSTALLED_PLUGINS,
    ADRExampleIssue,
    DetectedService,
    RegisteredMetadata,
)
from debputy.plugin.debputy.debputy_plugin import initialize_debputy_features
from debputy.substitution import SubstitutionImpl, VariableContext, Substitution
from debputy.util import package_cross_check_precheck

RegisteredPackagerProvidedFile.register(PackagerProvidedFileClassSpec)


@dataclasses.dataclass(frozen=True, slots=True)
class PackageProcessingContextTestProvider(PackageProcessingContext):
    binary_package: BinaryPackage
    binary_package_version: str
    related_udeb_package: Optional[BinaryPackage]
    related_udeb_package_version: Optional[str]
    accessible_package_roots: Callable[[], Sequence[Tuple[BinaryPackage, VirtualPath]]]


def _initialize_plugin_under_test(
    plugin_metadata: DebputyPluginMetadata,
    load_debputy_plugin: bool = True,
) -> "InitializedPluginUnderTest":
    feature_set = PluginProvidedFeatureSet()
    substitution = SubstitutionImpl(
        unresolvable_substitutions=frozenset(["SOURCE_DATE_EPOCH", "PACKAGE"]),
        variable_context=VariableContext(
            FSROOverlay.create_root_dir("debian", "debian"),
        ),
        plugin_feature_set=feature_set,
    )

    if load_debputy_plugin:
        debputy_plugin_metadata = plugin_metadata_for_debputys_own_plugin(
            initialize_debputy_features
        )
        # Load debputy's own plugin first, so conflicts with debputy's plugin are detected early
        debputy_provider = DebputyPluginInitializerProvider(
            debputy_plugin_metadata,
            feature_set,
            substitution,
        )
        debputy_provider.load_plugin()

    plugin_under_test_provider = DebputyPluginInitializerProvider(
        plugin_metadata,
        feature_set,
        substitution,
    )
    plugin_under_test_provider.load_plugin()

    return InitializedPluginUnderTestImpl(
        plugin_metadata.plugin_name,
        feature_set,
        substitution,
    )


def _auto_load_plugin_from_filename(
    py_test_filename: str,
) -> "InitializedPluginUnderTest":
    dirname, basename = os.path.split(py_test_filename)
    plugin_name = PLUGIN_TEST_SUFFIX.sub("", basename).replace("_", "-")

    test_location = os.environ.get("DEBPUTY_TEST_PLUGIN_LOCATION", "uninstalled")
    if test_location == "uninstalled":
        json_basename = f"{plugin_name}.json"
        json_desc_file = os.path.join(dirname, json_basename)
        if "/" not in json_desc_file:
            json_desc_file = f"./{json_desc_file}"

        if os.path.isfile(json_desc_file):
            return _initialize_plugin_from_desc(json_desc_file)

        json_desc_file_in = f"{json_desc_file}.in"
        if os.path.isfile(json_desc_file_in):
            return _initialize_plugin_from_desc(json_desc_file)
        raise FileNotFoundError(
            f"Cannot determine the plugin JSON metadata descriptor: Expected it to be"
            f" {json_desc_file} or {json_desc_file_in}"
        )

    if test_location == "installed":
        plugin_metadata = find_json_plugin([str(DEBPUTY_PLUGIN_ROOT_DIR)], plugin_name)
        return _initialize_plugin_under_test(plugin_metadata, load_debputy_plugin=True)

    raise ValueError(
        'Invalid or unsupported "DEBPUTY_TEST_PLUGIN_LOCATION" environment variable. It must be either'
        ' unset OR one of "installed", "uninstalled".'
    )


def initialize_plugin_under_test(
    *,
    plugin_desc_file: Optional[str] = None,
) -> "InitializedPluginUnderTest":
    """Load and initialize a plugin for testing it

    This method will load the plugin via plugin description, which is the method that `debputy` does at
     run-time (in contrast to `initialize_plugin_under_test_preloaded`, which bypasses this concrete part
     of the flow).

    :param plugin_desc_file: The plugin description file (`.json`) that describes how to load the plugin.
      If omitted, `debputy` will attempt to attempt the plugin description file based on the test itself.
      This works for "single-file" plugins, where the description file and the test are right next to
      each other.

      Note that the description file is *not* required to a valid version at this stage (e.g., "N/A" or
      "@PLACEHOLDER@") is fine. So you still use this method if you substitute in the version during
      build after running the tests. To support this flow, the file name can also end with `.json.in`
      (instead of `.json`).
    :return: The loaded plugin for testing
    """
    if plugin_desc_file is None:
        caller_file = inspect.stack()[1].filename
        return _auto_load_plugin_from_filename(caller_file)
    if DEBPUTY_TEST_AGAINST_INSTALLED_PLUGINS:
        raise RuntimeError(
            "Running the test against an installed plugin does not work when"
            " plugin_desc_file is provided. Please skip this test. You can "
            " import DEBPUTY_TEST_AGAINST_INSTALLED_PLUGINS and use that as"
            " conditional for this purpose."
        )
    return _initialize_plugin_from_desc(plugin_desc_file)


def _initialize_plugin_from_desc(
    desc_file: str,
) -> "InitializedPluginUnderTest":
    if not desc_file.endswith((".json", ".json.in")):
        raise ValueError("The plugin file must end with .json or .json.in")

    plugin_metadata = parse_json_plugin_desc(desc_file)

    return _initialize_plugin_under_test(plugin_metadata, load_debputy_plugin=True)


def initialize_plugin_under_test_from_inline_json(
    plugin_name: str,
    json_content: str,
) -> "InitializedPluginUnderTest":
    with BytesIO(json_content.encode("utf-8")) as fd:
        plugin_metadata = parse_json_plugin_desc(plugin_name, fd=fd)

    return _initialize_plugin_under_test(plugin_metadata, load_debputy_plugin=True)


def initialize_plugin_under_test_preloaded(
    api_compat_version: int,
    plugin_initializer: PluginInitializationEntryPoint,
    /,
    plugin_name: str = "plugin-under-test",
    load_debputy_plugin: bool = True,
    plugin_doc_path_resolver: Callable[
        [], Optional[Union[str, Traversable, Path]]
    ] = lambda: None,
) -> "InitializedPluginUnderTest":
    """Internal API: Initialize a plugin for testing without loading it from a file

    This method by-passes the standard loading mechanism, meaning you will not test that your plugin
    description file is correct. Notably, any feature provided via the JSON description file will
    **NOT** be visible for the test.

    This API is mostly useful for testing parts of debputy itself.

    :param api_compat_version: The API version the plugin was written for. Use the same version as the
      version from the entry point (The `v1` part of `debputy.plugins.v1.initialize` translate into `1`).
    :param plugin_initializer: The entry point of the plugin
    :param plugin_name: Normally, debputy would derive this from the entry point. In the test, it will
      use a test name and version. However, you can explicitly set if you want the real name/version.
    :param load_debputy_plugin: Whether to load debputy's own plugin first. Doing so provides a more
      realistic test and enables the test to detect conflicts with debputy's own plugins (de facto making
      the plugin unloadable in practice if such a conflict is present).  This option is mostly provided
      to enable debputy to use this method for self testing.
    :param plugin_doc_path_resolver: How to resolve the documentation (if relevant for the test). The
      default is to not load the documentation.
    :return: The loaded plugin for testing
    """

    if DEBPUTY_TEST_AGAINST_INSTALLED_PLUGINS:
        raise RuntimeError(
            "Running the test against an installed plugin does not work when"
            " the plugin is preload. Please skip this test. You can "
            " import DEBPUTY_TEST_AGAINST_INSTALLED_PLUGINS and use that as"
            " conditional for this purpose."
        )

    plugin_metadata = DebputyPluginMetadata(
        plugin_name=plugin_name,
        api_compat_version=api_compat_version,
        plugin_initializer=plugin_initializer,
        plugin_loader=None,
        plugin_path="<loaded-via-test>",
    )

    return _initialize_plugin_under_test(
        plugin_metadata,
        load_debputy_plugin=load_debputy_plugin,
    )


class _MockArchTable:
    @staticmethod
    def matches_architecture(_a: str, _b: str) -> bool:
        return True


FAKE_DPKG_QUERY_TABLE = cast("DpkgArchTable", _MockArchTable())
del _MockArchTable


def package_metadata_context(
    *,
    host_arch: str = "amd64",
    package_fields: Optional[Dict[str, str]] = None,
    related_udeb_package_fields: Optional[Dict[str, str]] = None,
    binary_package_version: str = "1.0-1",
    related_udeb_package_version: Optional[str] = None,
    should_be_acted_on: bool = True,
    related_udeb_fs_root: Optional[VirtualPath] = None,
    accessible_package_roots: Sequence[Tuple[Mapping[str, str], VirtualPath]] = tuple(),
) -> PackageProcessingContext:
    process_table = faked_arch_table(host_arch)
    f = {
        "Package": "foo",
        "Architecture": "any",
    }
    if package_fields is not None:
        f.update(package_fields)

    bin_package = BinaryPackage(
        Deb822(f),
        process_table,
        FAKE_DPKG_QUERY_TABLE,
        is_main_package=True,
        should_be_acted_on=should_be_acted_on,
    )
    udeb_package = None
    if related_udeb_package_fields is not None:
        uf = dict(related_udeb_package_fields)
        uf.setdefault("Package", f'{f["Package"]}-udeb')
        uf.setdefault("Architecture", f["Architecture"])
        uf.setdefault("Package-Type", "udeb")
        udeb_package = BinaryPackage(
            Deb822(uf),
            process_table,
            FAKE_DPKG_QUERY_TABLE,
            is_main_package=False,
            should_be_acted_on=True,
        )
        if related_udeb_package_version is None:
            related_udeb_package_version = binary_package_version
    if accessible_package_roots:
        apr = []
        for fields, apr_fs_root in accessible_package_roots:
            apr_fields = Deb822(dict(fields))
            if "Package" not in apr_fields:
                raise ValueError(
                    "Missing mandatory Package field in member of accessible_package_roots"
                )
            if "Architecture" not in apr_fields:
                raise ValueError(
                    "Missing mandatory Architecture field in member of accessible_package_roots"
                )
            apr_package = BinaryPackage(
                apr_fields,
                process_table,
                FAKE_DPKG_QUERY_TABLE,
                is_main_package=False,
                should_be_acted_on=True,
            )
            r = package_cross_check_precheck(bin_package, apr_package)
            if not r[0]:
                raise ValueError(
                    f"{apr_package.name} would not be accessible for {bin_package.name}"
                )
            apr.append((apr_package, apr_fs_root))

        if related_udeb_fs_root is not None:
            if udeb_package is None:
                raise ValueError(
                    "related_udeb_package_fields must be given when related_udeb_fs_root is given"
                )
            r = package_cross_check_precheck(bin_package, udeb_package)
            if not r[0]:
                raise ValueError(
                    f"{udeb_package.name} would not be accessible for {bin_package.name}, so providing"
                    " related_udeb_fs_root is irrelevant"
                )
            apr.append(udeb_package)
        apr = tuple(apr)
    else:
        apr = tuple()

    return PackageProcessingContextTestProvider(
        binary_package=bin_package,
        related_udeb_package=udeb_package,
        binary_package_version=binary_package_version,
        related_udeb_package_version=related_udeb_package_version,
        accessible_package_roots=lambda: apr,
    )


def manifest_variable_resolution_context(
    *,
    debian_dir: Optional[VirtualPath] = None,
) -> VariableContext:
    if debian_dir is None:
        debian_dir = FSRootDir()

    return VariableContext(debian_dir)


class MaintscriptAccessorTestProvider(MaintscriptAccessorProviderBase):
    __slots__ = ("_plugin_metadata", "_plugin_source_id", "_maintscript_container")

    def __init__(
        self,
        plugin_metadata: DebputyPluginMetadata,
        plugin_source_id: str,
        maintscript_container: Dict[str, List[RegisteredMaintscript]],
    ):
        self._plugin_metadata = plugin_metadata
        self._plugin_source_id = plugin_source_id
        self._maintscript_container = maintscript_container

    @classmethod
    def _apply_condition_to_script(
        cls, condition: str, run_snippet: str, /, indent: Optional[bool] = None
    ) -> str:
        return run_snippet

    def _append_script(
        self,
        caller_name: str,
        maintscript: Maintscript,
        full_script: str,
        /,
        perform_substitution: bool = True,
    ) -> None:
        if self._plugin_source_id not in self._maintscript_container:
            self._maintscript_container[self._plugin_source_id] = []
        self._maintscript_container[self._plugin_source_id].append(
            RegisteredMaintscript(
                maintscript,
                caller_name,
                full_script,
                perform_substitution,
            )
        )


class RegisteredMetadataImpl(RegisteredMetadata):
    __slots__ = (
        "_substvars",
        "_triggers",
        "_maintscripts",
    )

    def __init__(
        self,
        substvars: Substvars,
        triggers: List[RegisteredTrigger],
        maintscripts: List[RegisteredMaintscript],
    ) -> None:
        self._substvars = substvars
        self._triggers = triggers
        self._maintscripts = maintscripts

    @property
    def substvars(self) -> Substvars:
        return self._substvars

    @property
    def triggers(self) -> List[RegisteredTrigger]:
        return self._triggers

    def maintscripts(
        self,
        *,
        maintscript: Optional[Maintscript] = None,
    ) -> List[RegisteredMaintscript]:
        if maintscript is None:
            return self._maintscripts
        return [m for m in self._maintscripts if m.maintscript == maintscript]


class BinaryCtrlAccessorTestProvider(BinaryCtrlAccessorProviderBase):
    __slots__ = ("_maintscript_container",)

    def __init__(
        self,
        plugin_metadata: DebputyPluginMetadata,
        plugin_source_id: str,
        context: PackageProcessingContext,
    ) -> None:
        super().__init__(
            plugin_metadata,
            plugin_source_id,
            context,
            {},
            FlushableSubstvars(),
            (None, None),
        )
        self._maintscript_container: Dict[str, List[RegisteredMaintscript]] = {}

    def _create_maintscript_accessor(self) -> MaintscriptAccessor:
        return MaintscriptAccessorTestProvider(
            self._plugin_metadata,
            self._plugin_source_id,
            self._maintscript_container,
        )

    def registered_metadata(self) -> RegisteredMetadata:
        return RegisteredMetadataImpl(
            self._substvars,
            [
                RegisteredTrigger.from_plugin_provided_trigger(t)
                for t in self._triggers.values()
                if t.provider_source_id == self._plugin_source_id
            ],
            self._maintscript_container.get(self._plugin_source_id, []),
        )


class ServiceRegistryTestImpl(ServiceRegistry[DSD]):
    __slots__ = ("_service_manager_details", "_service_definitions")

    def __init__(
        self,
        service_manager_details: ServiceManagerDetails,
        detected_services: List[DetectedService[DSD]],
    ) -> None:
        self._service_manager_details = service_manager_details
        self._service_definitions = detected_services

    def register_service(
        self,
        path: VirtualPath,
        name: Union[str, List[str]],
        *,
        type_of_service: str = "service",  # "timer", etc.
        service_scope: str = "system",
        enable_by_default: bool = True,
        start_by_default: bool = True,
        default_upgrade_rule: ServiceUpgradeRule = "restart",
        service_context: Optional[DSD] = None,
    ) -> None:
        names = name if isinstance(name, list) else [name]
        if len(names) < 1:
            raise ValueError(
                f"The service must have at least one name - {path.absolute} did not have any"
            )
        self._service_definitions.append(
            DetectedService(
                path,
                names,
                type_of_service,
                service_scope,
                enable_by_default,
                start_by_default,
                default_upgrade_rule,
                service_context,
            )
        )


@contextlib.contextmanager
def _read_only_fs_root(fs_root: VirtualPath) -> Iterator[VirtualPath]:
    if fs_root.is_read_write:
        assert isinstance(fs_root, FSRootDir)
        fs_root.is_read_write = False
        yield fs_root
        fs_root.is_read_write = True
    else:
        yield fs_root


class InitializedPluginUnderTestImpl(InitializedPluginUnderTest):
    def __init__(
        self,
        plugin_name: str,
        feature_set: PluginProvidedFeatureSet,
        substitution: SubstitutionImpl,
    ) -> None:
        self._feature_set = feature_set
        self._plugin_name = plugin_name
        self._packager_provided_files: Optional[
            Dict[str, RegisteredPackagerProvidedFile]
        ] = None
        self._triggers: Dict[Tuple[DpkgTriggerType, str], PluginProvidedTrigger] = {}
        self._maintscript_container: Dict[str, List[RegisteredMaintscript]] = {}
        self._substitution = substitution
        assert plugin_name in self._feature_set.plugin_data

    @property
    def _plugin_metadata(self) -> DebputyPluginMetadata:
        return self._feature_set.plugin_data[self._plugin_name]

    def packager_provided_files_by_stem(
        self,
    ) -> Mapping[str, RegisteredPackagerProvidedFile]:
        ppf = self._packager_provided_files
        if ppf is None:
            result: Dict[str, RegisteredPackagerProvidedFile] = {}
            for spec in self._feature_set.packager_provided_files.values():
                if spec.debputy_plugin_metadata.plugin_name != self._plugin_name:
                    continue
                # Registered as a virtual subclass, so this should always be True
                assert isinstance(spec, RegisteredPackagerProvidedFile)
                result[spec.stem] = spec
            self._packager_provided_files = result
            ppf = result
        return ppf

    def run_metadata_detector(
        self,
        metadata_detector_id: str,
        fs_root: VirtualPath,
        context: Optional[PackageProcessingContext] = None,
    ) -> RegisteredMetadata:
        if fs_root.parent_dir is not None:
            raise ValueError("Provided path must be the file system root.")
        detectors = self._feature_set.metadata_maintscript_detectors[self._plugin_name]
        matching_detectors = [
            d for d in detectors if d.detector_id == metadata_detector_id
        ]
        if len(matching_detectors) != 1:
            assert not matching_detectors
            raise ValueError(
                f"The plugin {self._plugin_name} did not provide a metadata detector with ID"
                f' "{metadata_detector_id}"'
            )
        if context is None:
            context = package_metadata_context()
        detector = matching_detectors[0]
        if not detector.applies_to(context.binary_package):
            raise ValueError(
                f'The detector "{metadata_detector_id}" from {self._plugin_name} does not apply to the'
                " given package. Consider using `package_metadata_context()` to emulate a binary package"
                " with the correct specification. As an example: "
                '`package_metadata_context(package_fields={"Package-Type": "udeb"})` would emulate a udeb'
                " package."
            )

        ctrl = BinaryCtrlAccessorTestProvider(
            self._plugin_metadata,
            metadata_detector_id,
            context,
        )
        with _read_only_fs_root(fs_root) as ro_root:
            detector.run_detector(
                ro_root,
                ctrl,
                context,
            )
        return ctrl.registered_metadata()

    def run_package_processor(
        self,
        package_processor_id: str,
        fs_root: VirtualPath,
        context: Optional[PackageProcessingContext] = None,
    ) -> None:
        if fs_root.parent_dir is not None:
            raise ValueError("Provided path must be the file system root.")
        pp_key = (self._plugin_name, package_processor_id)
        package_processor = self._feature_set.all_package_processors.get(pp_key)
        if package_processor is None:
            raise ValueError(
                f"The plugin {self._plugin_name} did not provide a package processor with ID"
                f' "{package_processor_id}"'
            )
        if context is None:
            context = package_metadata_context()
        if not fs_root.is_read_write:
            raise ValueError(
                "The provided fs_root is read-only and it must be read-write for package processor"
            )
        if not package_processor.applies_to(context.binary_package):
            raise ValueError(
                f'The package processor "{package_processor_id}" from {self._plugin_name} does not apply'
                " to the given package. Consider using `package_metadata_context()` to emulate a binary"
                " package with the correct specification. As an example: "
                '`package_metadata_context(package_fields={"Package-Type": "udeb"})` would emulate a udeb'
                " package."
            )
        package_processor.run_package_processor(
            fs_root,
            None,
            context,
        )

    @property
    def declared_manifest_variables(self) -> FrozenSet[str]:
        return frozenset(
            {
                k
                for k, v in self._feature_set.manifest_variables.items()
                if v.plugin_metadata.plugin_name == self._plugin_name
            }
        )

    def automatic_discard_rules_examples_with_issues(self) -> Sequence[ADRExampleIssue]:
        issues = []
        for adr in self._feature_set.auto_discard_rules.values():
            if adr.plugin_metadata.plugin_name != self._plugin_name:
                continue
            for idx, example in enumerate(adr.examples):
                result = process_discard_rule_example(
                    adr,
                    example,
                )
                if result.inconsistent_paths:
                    issues.append(
                        ADRExampleIssue(
                            adr.name,
                            idx,
                            [
                                x.absolute + ("/" if x.is_dir else "")
                                for x in result.inconsistent_paths
                            ],
                        )
                    )
        return issues

    def run_service_detection_and_integrations(
        self,
        service_manager: str,
        fs_root: VirtualPath,
        context: Optional[PackageProcessingContext] = None,
        *,
        service_context_type_hint: Optional[Type[DSD]] = None,
    ) -> Tuple[List[DetectedService[DSD]], RegisteredMetadata]:
        if fs_root.parent_dir is not None:
            raise ValueError("Provided path must be the file system root.")
        try:
            service_manager_details = self._feature_set.service_managers[
                service_manager
            ]
            if service_manager_details.plugin_metadata.plugin_name != self._plugin_name:
                raise KeyError(service_manager)
        except KeyError:
            raise ValueError(
                f"The plugin {self._plugin_name} does not provide a"
                f" service manager called {service_manager}"
            ) from None

        if context is None:
            context = package_metadata_context()
        detected_services: List[DetectedService[DSD]] = []
        registry = ServiceRegistryTestImpl(service_manager_details, detected_services)
        service_manager_details.service_detector(
            fs_root,
            registry,
            context,
        )
        ctrl = BinaryCtrlAccessorTestProvider(
            self._plugin_metadata,
            service_manager_details.service_manager,
            context,
        )
        if detected_services:
            service_definitions = [
                ServiceDefinitionImpl(
                    ds.names[0],
                    ds.names,
                    ds.path,
                    ds.type_of_service,
                    ds.service_scope,
                    ds.enable_by_default,
                    ds.start_by_default,
                    ds.default_upgrade_rule,
                    self._plugin_name,
                    True,
                    ds.service_context,
                )
                for ds in detected_services
            ]
            service_manager_details.service_integrator(
                service_definitions,
                ctrl,
                context,
            )
        return detected_services, ctrl.registered_metadata()

    def manifest_variables(
        self,
        *,
        resolution_context: Optional[VariableContext] = None,
        mocked_variables: Optional[Mapping[str, str]] = None,
    ) -> Mapping[str, str]:
        valid_manifest_variables = frozenset(
            {
                n
                for n, v in self._feature_set.manifest_variables.items()
                if v.plugin_metadata.plugin_name == self._plugin_name
            }
        )
        if resolution_context is None:
            resolution_context = manifest_variable_resolution_context()
        substitution = self._substitution.copy_for_subst_test(
            self._feature_set,
            resolution_context,
            extra_substitutions=mocked_variables,
        )
        return SubstitutionTable(
            valid_manifest_variables,
            substitution,
        )


class SubstitutionTable(Mapping[str, str]):
    def __init__(
        self, valid_manifest_variables: FrozenSet[str], substitution: Substitution
    ) -> None:
        self._valid_manifest_variables = valid_manifest_variables
        self._resolved: Set[str] = set()
        self._substitution = substitution

    def __contains__(self, item: object) -> bool:
        return item in self._valid_manifest_variables

    def __getitem__(self, key: str) -> str:
        if key not in self._valid_manifest_variables:
            raise KeyError(key)
        v = self._substitution.substitute(
            "{{" + key + "}}", f"test of manifest variable `{key}`"
        )
        self._resolved.add(key)
        return v

    def __len__(self) -> int:
        return len(self._valid_manifest_variables)

    def __iter__(self) -> Iterator[str]:
        return iter(self._valid_manifest_variables)

    def keys(self) -> KeysView[str]:
        return cast("KeysView[str]", self._valid_manifest_variables)



import dataclasses
import re
from typing import Sequence, Optional, FrozenSet, Tuple, List, cast

from debputy.architecture_support import DpkgArchitectureBuildProcessValuesTable
from debputy.commands.debputy_cmd.output import OutputStylingBase
from debputy.highlevel_manifest import MutableYAMLManifest
from debputy.substitution import Substitution

_DH_VAR_RE = re.compile(r"([$][{])([A-Za-z0-9][-_:0-9A-Za-z]*)([}])")


class AcceptableMigrationIssues:
    def __init__(self, values: FrozenSet[str]):
        self._values = values

    def __contains__(self, item: str) -> bool:
        return item in self._values or "ALL" in self._values


class UnsupportedFeature(RuntimeError):
    @property
    def message(self) -> str:
        return cast("str", self.args[0])

    @property
    def issue_keys(self) -> Optional[Sequence[str]]:
        if len(self.args) < 2:
            return None
        return cast("Sequence[str]", self.args[1])


class ConflictingChange(RuntimeError):
    @property
    def message(self) -> str:
        return cast("str", self.args[0])


@dataclasses.dataclass(slots=True)
class FeatureMigration:
    tagline: str
    fo: OutputStylingBase
    successful_manifest_changes: int = 0
    already_present: int = 0
    warnings: List[str] = dataclasses.field(default_factory=list)
    remove_paths_on_success: List[str] = dataclasses.field(default_factory=list)
    rename_paths_on_success: List[Tuple[str, str]] = dataclasses.field(
        default_factory=list
    )
    assumed_compat: Optional[int] = None
    required_plugins: List[str] = dataclasses.field(default_factory=list)

    def warn(self, msg: str) -> None:
        self.warnings.append(msg)

    def rename_on_success(self, source: str, dest: str) -> None:
        self.rename_paths_on_success.append((source, dest))

    def remove_on_success(self, path: str) -> None:
        self.remove_paths_on_success.append(path)

    def require_plugin(self, debputy_plugin: str) -> None:
        self.required_plugins.append(debputy_plugin)

    @property
    def anything_to_do(self) -> bool:
        return bool(self.total_changes_involved)

    @property
    def performed_changes(self) -> int:
        return (
            self.successful_manifest_changes
            + len(self.remove_paths_on_success)
            + len(self.rename_paths_on_success)
        )

    @property
    def total_changes_involved(self) -> int:
        return (
            self.successful_manifest_changes
            + len(self.warnings)
            + len(self.remove_paths_on_success)
            + len(self.rename_paths_on_success)
        )


class DHMigrationSubstitution(Substitution):
    def __init__(
        self,
        dpkg_arch_table: DpkgArchitectureBuildProcessValuesTable,
        acceptable_migration_issues: AcceptableMigrationIssues,
        feature_migration: FeatureMigration,
        mutable_manifest: MutableYAMLManifest,
    ) -> None:
        self._acceptable_migration_issues = acceptable_migration_issues
        self._dpkg_arch_table = dpkg_arch_table
        self._feature_migration = feature_migration
        self._mutable_manifest = mutable_manifest
        # TODO: load 1:1 variables from the real subst instance (less stuff to keep in sync)
        one2one = [
            "DEB_SOURCE",
            "DEB_VERSION",
            "DEB_VERSION_EPOCH_UPSTREAM",
            "DEB_VERSION_UPSTREAM_REVISION",
            "DEB_VERSION_UPSTREAM",
            "SOURCE_DATE_EPOCH",
        ]
        self._builtin_substs = {
            "Tab": "{{token:TAB}}",
            "Space": " ",
            "Newline": "{{token:NEWLINE}}",
            "Dollar": "${}",
        }
        self._builtin_substs.update((x, "{{" + x + "}}") for x in one2one)

    def _replacement(self, key: str, definition_source: str) -> str:
        if key in self._builtin_substs:
            return self._builtin_substs[key]
        if key in self._dpkg_arch_table:
            return "{{" + key + "}}"
        if key.startswith("env:"):
            if "dh-subst-env" not in self._acceptable_migration_issues:
                raise UnsupportedFeature(
                    "Use of environment based substitution variable {{"
                    + key
                    + "}} is not"
                    f" supported in debputy. The variable was spotted at {definition_source}",
                    ["dh-subst-env"],
                )
        elif "dh-subst-unknown-variable" not in self._acceptable_migration_issues:
            raise UnsupportedFeature(
                "Unknown substitution variable {{"
                + key
                + "}}, which does not have a known"
                f" counter part in debputy. The variable was spotted at {definition_source}",
                ["dh-subst-unknown-variable"],
            )
        manifest_definitions = self._mutable_manifest.manifest_definitions(
            create_if_absent=False
        )
        manifest_variables = manifest_definitions.manifest_variables(
            create_if_absent=False
        )
        if key not in manifest_variables.variables:
            manifest_definitions.create_definition_if_missing()
            manifest_variables[key] = "TODO: Provide variable value for " + key
            self._feature_migration.warn(
                "TODO: MANUAL MIGRATION of unresolved substitution variable {{"
                + key
                + "}} from"
                + f" {definition_source}"
            )
            self._feature_migration.successful_manifest_changes += 1

        return "{{" + key + "}}"

    def substitute(
        self,
        value: str,
        definition_source: str,
        /,
        escape_glob_characters: bool = False,
    ) -> str:
        if "${" not in value:
            return value
        replacement = self._apply_substitution(
            _DH_VAR_RE,
            value,
            definition_source,
            escape_glob_characters=escape_glob_characters,
        )
        return replacement.replace("${}", "$")

    def with_extra_substitutions(self, **extra_substitutions: str) -> "Substitution":
        return self



import textwrap

import pytest

from debputy.lsp.debputy_ls import DebputyLanguageServer
from debputy.lsprotocol.types import (
    TextDocumentIdentifier,
    SemanticTokensParams,
)

try:
    from debputy.lsp.lsp_debian_patches_series import (
        _debian_patches_semantic_tokens_full,
        _debian_patches_series_completions,
    )

    from pygls.server import LanguageServer
except ImportError:
    pass
from lsp_tests.lsp_tutil import (
    put_doc_no_cursor,
    resolve_semantic_tokens,
    resolved_semantic_token,
)


@pytest.mark.asyncio
async def test_dpatches_series_semantic_tokens(ls: "DebputyLanguageServer") -> None:
    doc_uri = "file:///nowhere/debian/patches/series"
    content = textwrap.dedent(
        """\
        # Some leading comment

        some.patch

        another-delta.diff # foo
"""
    )
    put_doc_no_cursor(
        ls,
        doc_uri,
        "debian/patches/series",
        content,
    )

    lines = content.splitlines(keepends=True)
    semantic_tokens = await _debian_patches_semantic_tokens_full(
        ls,
        SemanticTokensParams(TextDocumentIdentifier(doc_uri)),
    )
    resolved_semantic_tokens = resolve_semantic_tokens(lines, semantic_tokens)
    assert resolved_semantic_tokens is not None
    assert resolved_semantic_tokens == [
        resolved_semantic_token(0, 0, "# Some leading comment", "comment"),
        resolved_semantic_token(2, 0, "some.patch", "string"),
        resolved_semantic_token(4, 0, "another-delta.diff", "string"),
        resolved_semantic_token(4, len("another-delta.diff") + 1, "# foo", "comment"),
    ]
